"""
Example callbacks (:mod:`stnmf.callbacks`)
==========================================

Exemplary callbacks illustrating their use with :class:`stnmf.STNMF`.

.. autosummary::
    :toctree: generated/

    consensus
    perturbation
    residual
    stopping_criterion

Detail
------
Callbacks allow to safely modify the algorithm of subunit decomposition
in many aspects without having to dive deep into the NMF code. The user
can specify a function that will be called at each loop iteration in the
iterative process of NMF. Code execution jumps out of the internal code
into the user function and returns afterwards to continue until the next
iteration. The function defined by the user will be supplied with and
has access to the current (mutable) state of the decomposition at each
iteration.

Among many other things, this allows...

* to record the course of certain properties throughout the iterative
  process,

* to have the process be interrupted prematurely based on termination
  criteria,

* to change properties of the current decomposition to modify the course
  of the decomposition,

* and to leverage parts of the algorithm to completely redefine how
  subunits are decomposed.

A callback function can be understood as a plug-in or add-on extension
of the internal code.

"""
__all__ = [
    'consensus',
    'perturbation',
    'residual',
    'stopping_criterion',
]

from .consensus import callback as consensus
from .perturbation import callback as perturbation
from .residual import callback as residual
from .stopping_criterion import callback as stopping_criterion



"""
Spike-triggered analyses (:mod:`stnmf.preprocessing.spiketriggered`)
====================================================================

Collection of functions performing spike-triggered analyses for
multiple cells

.. autosummary::
    :toctree: generated/

    average
    stimulusensemble
    window
    profiles

Data preparation
----------------
:class:`stnmf.STNMF` expects an effective spike-triggered stimulus
ensemble (STE) as input. To format the stimulus and spike response into
an STE the following pipeline can be utilized.

Data requirements
^^^^^^^^^^^^^^^^^

* **Spatiotemporal stimulus** as iterator or iterable with each element
  yielding a `(x,y,trial_length)` :term:`array_like`. For instance,
  that may be a `(num_trials,x,y,trial_length)` :class:`numpy.ndarray`.
  The spatial dimensions should enclose the entire projector screen. It
  will be cropped for each cell. Each array element is a pixel contrast
  value deviating from zero either negatively (darker) or positively
  (brighter).

* **Binned spike counts** as `(num_trials,trial_length,num_cells)`
  :term:`array_like`. Each array element is a non-negative integer
  describing the number of spikes within the (equally sized) time bin.

Pipeline
^^^^^^^^

   >>> from stnmf.preprocessing import spiketriggered

1. Calculate the spike-triggered averages (STAs) of all cells.

   >>> stas = spiketriggered.average(stimulus, spikes, tau)

2. Find the relevant spatial windows around the STAs for all cells.

   >>> crops = spiketriggered.window(stas)

3. Create the temporal profile of the STAs.

   >>> sta_space, sta_temp = spiketriggered.profiles(stas, crops)

4. For each cell `idx` individually, select its respective data

   >>> stim = stimulus[:, crops[idx], :]
   >>> spk = spikes[..., idx]
   >>> temporal = sta_temp[idx]

   and create the STE, the input for STNMF.

   >>> ste = spiketriggered.stimulusensemble(stim, spk, temporal)

.. note::
    Review the individual function signatures to specify additional
    parameters.

"""
import numpy as np

from .. import spatial
from ..util import tqdm

__all__ = [
    'average',
    'stimulusensemble',
    'window',
    'profiles',
]


def average(stimulus, spikes, tau, continuous=False, tqdm_args=None):
    """
    Compute the spike-triggered average (STA) for multiple cells

    Parameters
    ----------
    stimulus : iterator or array_like
        This iterator should yield a `(x,y,trial_length)` array_like
        for each trial, where `trial_length` matches the number
        of elements of `spikes`

    spikes : (num_trials,trial_length,num_cells) array_like
        Binned spike counts of multiple cells. If two dimensional,
        assume `num_cells == 1`

    tau : int
        Temporal filter length

    continuous : bool, optional
        Specify if trials should be treated as connected chunks of a
        continuous recording or if they are separated. Default is False

    tqdm_args : dict, optional
        Arguments to pass to tqdm progress bar. If empty, tqdm is not
        used. Default is None

    Return
    ------
    sta : (num_cells,x,y,tau) numpy.ndarray
        STAs of all cells
    """
    spikes = np.asarray(spikes, dtype='float32')
    if spikes.ndim == 2:
        spikes = spikes[..., None]
    num_trials, num_frames, num_cells = spikes.shape

    # Excluding first tau spikes
    if not continuous:
        spikes = spikes[:, tau-1:, :]
        sum_spikes = spikes.sum(axis=(0, 1))
    else:
        sum_spikes = spikes.reshape(-1, num_cells)[tau-1:].sum(axis=0)

    sta, sts = None, None
    it_spikes = tqdm(spikes, **tqdm_args) if tqdm_args else spikes
    for sp, stim in zip(it_spikes, stimulus):
        # Allocate arrays once on first iteration
        if sta is None:
            x, y = stim.shape[:-1]
            shape = (x*y, tau, num_cells)
            sta = np.zeros(shape, dtype='float32')
            sts = np.zeros(shape, dtype='float32')
            if continuous:
                sp = sp[tau-1:]  # Only on first chunk
                stim_prev = stim[..., -(tau-1):]
        elif continuous:
            stim = np.concatenate((stim_prev, stim), axis=-1)
            stim_prev = stim[..., -(tau-1):]
            num_frames = stim.shape[-1]

        stim = stim.reshape(-1, num_frames).astype('float32')
        for it in range(tau):
            np.matmul(stim[:, it:num_frames-tau+it+1], sp, out=sts[:, it])
        sta += sts

    with np.errstate(invalid='ignore', divide='ignore'):
        sta = np.divide(sta, sum_spikes, dtype='float64')
    sta = np.moveaxis(sta, -1, 0)
    sta = sta.reshape(num_cells, x, y, tau)

    return sta


def window(sta, sd=3):
    """
    Obtain the spatial crop dimensions containing the receptive field
    for multiple cells

    Parameters
    ----------
    sta : (num_cells,x,y,tau) array_like
        STAs of multiple cells. If three dimensional, assume
        `num_cells == 1`

    sd : float, optional
        Standard deviations of a two-dimensional Gaussian fit for the
        spatial windows around the center. Default is 3

    Returns
    -------
    crop : (num_cells,) numpy.ndarray containing tuple of slice
        Object array containing the spatial crop dimensions for all
        cells. Each element is a tuple of slices with
        `(slice(x_min,x_max,None), slice(y_min,y_max,None))`
    """
    sta = np.asarray(sta)
    if sta.ndim == 3:
        sta = sta[None, ...]
    num_cells, x, y, tau = sta.shape

    crop = np.empty(num_cells, dtype='object')
    for i, s in enumerate(sta):
        # Determine preliminary spatial crop
        cropsize = 20
        ind_max = np.abs(s).argmax()
        x_max, y_max, t_max = np.unravel_index(ind_max, s.shape)
        window = np.tile([x_max, y_max], [2, 1]).T + [-cropsize, cropsize+1]
        x_low, y_low, x_high, y_high = np.clip(window.T, 0, (x, y)).reshape(-1)

        # Correct sign, flip spatial STA to positive sign for ellipse fitting
        s = s[x_low:x_high, y_low:y_high, :]
        sta_space = s[..., t_max]
        if -sta_space.min() > sta_space.max():
            sta_space = -sta_space

        # Fit two-dimensional Gaussian and determine crop window
        mux, muy, r_major, r_minor, rad = spatial.ellipse(sta_space, sigma=sd)
        uxy = [r_major, r_minor] * np.cos([rad, rad+np.pi/2])
        vxy = [r_major, r_minor] * np.sin([rad, rad+np.pi/2])
        halfsize = np.linalg.norm([uxy, vxy], axis=1)
        extent = np.int32([[mux, muy] - halfsize, [mux, muy] + halfsize+1])
        extent += [x_low, y_low]  # Add trimmed region for whole screen
        crop[i] = (slice(*extent[:, 0]), slice(*extent[:, 1]))  # Form slices

    return crop


def profiles(sta, crop=None):
    """
    Obtain temporal and spatial profiles of the STA for multiple cells

    Parameters
    ----------
    sta : (num_cells,x,y,tau) array_like
        STAs of multiple cells. If three dimensional, assume
        `num_cells == 1`

    crop : list of tuple of slice, optional
        Windows for the spatial profiles. Output of the `window`
        function. Default is None

    Returns
    -------
    sta_space : (num_cells,) numpy.ndarray
        Object array containing the spatial filter for all cells. Each
        element is a `(crop_x[cell],crop_y[cell],tau)` numpy.ndarray,
        if `crop` is provided, otherwise a `(x,y,tau)` numpy.ndarray

    sta_temp : (num_cells,tau) numpy.ndarray
        Array containing the normalized temporal filters for all cells
    """
    sta = np.asarray(sta)
    if sta.ndim == 3:
        sta = sta[None, ...]
    num_cells, x, y, tau = sta.shape

    if crop is None:
        crop = np.full(num_cells, 0, dtype='O')
        crop[:] = [tuple([Ellipsis])]

    sta_space = np.empty(num_cells, 'O')
    sta_temp = np.empty((num_cells, tau), 'float32')
    for i, s in enumerate(sta):
        ind_max = np.abs(s).argmax()
        x_max, y_max, t_max = np.unravel_index(ind_max, s.shape)
        sta_temp[i] = s[x_max, y_max]
        sta_temp[i] /= np.linalg.norm(sta_temp[i])
        sta_space[i] = s[crop[i] + (t_max,)]

    return sta_space, sta_temp


def stimulusensemble(stimulus, spikes, sta_temp, continuous=False,
                     min_spikes_per_bin=1, dtype='float32'):
    """
    Collect the effective spike-triggered stimulus ensemble (STE)

    Parameters
    ----------
    stimulus : iterator or array_like
        This iterator should yield a `(x,y,trial_length)` array_like
        for each trial, where `trial_length` matches the number
        of elements of `spikes`. The stimulus should be cropped around
        the cell's receptive field. See function `window`

    spikes : (num_trials,trial_length,num_cells) array_like
        Binned spike counts of one cell, i.e. `num_cells` has to be 1
        (see notes). If two dimensional, assume `num_cells == 1`

    sta_temp : (tau,) array_like
        Temporal profile of the spike-triggered average. See function
        `profiles`

    continuous : bool, optional
        Specify if trials should be treated as connected chunks of a
        continuous recording or if they are separated. Default is False

    min_spikes_per_bin : int, optional
        Specify the minimum number of spikes per bin to be considered.
        This is useful for cells with high baseline firing rate. Default
        is 1

    dtype : str or numpy.dtype, optional
        Data type in which to perform and return the computations.
        Default is 'float32'

    Returns
    -------
    ste : (x,y,s) numpy.ndarray
        Effective Spike-triggered stimulus ensemble, where `s` is the
        number of spikes

    Raises
    ------
    NotImplementedError :
        If the last dimension of the spikes array is greater than one

    Notes
    -----
    The current implementation only supports spikes from one cell at a
    time.

    Each spike yields one spike-triggered stimulus, that is, multiple
    spikes in one bin result in identical successive spike-triggered
    stimulus frames in the STE.
    """
    sta_temp = np.asarray(sta_temp, dtype=dtype)
    tau = sta_temp.size

    spikes = np.asarray(spikes, dtype='int8')
    if spikes.ndim == 2:
        spikes = spikes[..., None]
    if spikes.shape[-1] != 1:
        raise NotImplementedError('The current implementation only '
                                  'supports spikes from one cell at a '
                                  'time.')

    # Minimum number of spikes to consider
    spikes = np.maximum(spikes - (min_spikes_per_bin-1), 0)

    # Excluding first tau spikes
    if not continuous:
        spikes = spikes[:, tau-1:, :]
        sum_spikes = spikes.sum()
    else:
        sum_spikes = spikes.reshape(-1)[tau-1:].sum()

    ste, sts = None, None
    s_ind = 0
    for sp, stim in zip(spikes, stimulus):
        # Allocate arrays once on first iteration
        if ste is None:
            shape = stim.shape[:2]
            sts = np.empty(shape + (1,), dtype=dtype)
            ste = np.empty(shape + (sum_spikes,), dtype=dtype)
            if continuous:
                sp = sp[tau-1:]  # Only on first chunk
                stim_prev = stim[..., -(tau-1):]
        elif continuous:
            stim = np.concatenate((stim_prev, stim), axis=-1)
            stim_prev = stim[..., -(tau-1):]

        ind = np.nonzero(sp)[0]
        for i in ind:
            np.dot(stim[..., i:i+tau], sta_temp, out=sts[..., 0])
            num_sp = int(sp[i, 0])
            np.copyto(ste[..., s_ind:s_ind+num_sp], sts)
            s_ind += num_sp

    # Flip sign to match polarity
    sign = np.argmax([sta_temp.argmin(), -np.inf, sta_temp.argmax()]) - 1
    ste *= sign

    return ste



"""
Spatial operations (:mod:`stnmf.spatial`)
=========================================

Collection of functions to perform spatial operations on
(two-dimensional) image arrays

.. autosummary::
    :toctree: generated/

    moransi
    ellipse
    contour

"""
import numpy as np
from scipy import ndimage
from scipy.optimize import leastsq
from scipy.signal import convolve
from scipy.stats import norm as scinorm
import shapely
from skimage import measure
import warnings

__all__ = [
    'moransi',
    'ellipse',
    'contour',
]


def moransi(image, p_value=False):
    """
    Compute Moran's I (spatial autocorrelation)

    Based on Moran's I [1]_ [2]_.

    Parameters
    ----------
    image : (sx,sy) array_like, or (r,sx,sy) array_like
        Two-dimensional image, or stack of `r` two-dimensional images

    p_value : bool, optional
        Return the p-value (significance) of the autocorrelation.
        Default is False

    Returns
    -------
    i : float, or (r,) numpy.ndarray
        Moran's I of the `image` array(s)

    p_value : float, or (r,) numpy.ndarray
        Associated p-value indicating the significance of
        autocorrelation. If `p_value` (function argument) is True

    Raises
    ------
    ValueError
        If `image` is neither two- nor three-dimensional.

    References
    ----------
    .. [1] Moran, P. A. P. (1950). Notes on Continuous Stochastic
           Phenomena. Biometrika. 37 (1): 17–23.
           https://doi.org/10.1093/BIOMET/37.1-2.17

    .. [2] Li, Hongfei; Calder, Catherine A.; Cressie, Noel (2007).
           Beyond Moran's I: Testing for Spatial Dependence Based on
           the Spatial Autoregressive Model. Geographical Analysis. 39
           (4): 357–375.
           https://doi.org/10.1111/j.1538-4632.2007.00708.x
    """
    ndim = np.ndim(image)
    if ndim not in [2, 3]:
        raise ValueError('image is expected to be two- or three-dimensional')

    # Mean of empty slice
    if np.size(image) == 0:
        if ndim == 3:
            i = np.zeros_like(image, shape=np.shape(image)[0])
            return (i, np.full_like(i, np.nan)) if p_value else i
        else:
            return (0, 0) if p_value else 0

    # Copy and fill NaNs
    s = np.nan_to_num(image, copy=True)

    sx, sy = s.shape[-2:]
    n = sx * sy
    n_edge = np.max(2 * (sx + sy) - 4, 0)

    # Subtract total mean from each element
    s -= s.mean(axis=(-2, -1), keepdims=True)
    sum_squared = (s**2).sum(axis=(-2, -1))
    sum_weights = 4 * n  # Each element has four weights
    sum_weights -= n_edge + 4  # Edge elements have no neighbors

    # Binary filter defining the neighbors
    filt = np.array([[0, 1, 0],
                     [1, 0, 1],
                     [0, 1, 0]], dtype=s.dtype)  # Type for performance
    if ndim == 3:
        filt = filt[None, ...]

    # Multiply each element with the sum of its neighbors
    neighbors_prod = s * convolve(s, filt, mode='same')
    neighbors_prod_sum = neighbors_prod.sum(axis=(-2, -1))

    with np.errstate(invalid='ignore'):
        i = n * neighbors_prod_sum / sum_weights / sum_squared

    if not p_value:
        return i

    # Z-score
    n_squared = n**2
    sum_weights_squared = sum_weights ** 2
    s1 = sum_weights * 2
    s2 = 4*16 + (n_edge-4)*36 + (n-n_edge)*64
    with np.errstate(invalid='ignore'):
        s3 = (s**4).sum(axis=(-2, -1)) / (1/n * sum_squared**2)
    s4 = (n_squared - 3*n + 3) * s1 - n*s2 + 3*sum_weights_squared
    s5 = (n_squared - n) * s1 - 2*n*s2 + 6*sum_weights_squared

    mu = -1 / (n - 1)
    var = (n*s4 - s3*s5) / ((n-1)*(n-2)*(n-3) * sum_weights_squared) - mu**2
    z_score = (i - mu) / np.sqrt(var)

    # P-value
    p_value = scinorm.sf(np.abs(z_score))*2

    return i, p_value


def ellipse(image, sigma=2.0):
    """
    Fit a two-dimensional Gaussian to an image and return a
    parameterized ellipse

    Parameters
    ----------
    image : (x,y) or (r,x,y) array_like
        Two-dimensional image or stack of `r` independent
        two-dimensional images, where `x` and `y` are the spatial
        dimensions

    sigma : float, optional
        Standard deviations (s.d.) of the mean (Mahalanobis distance)
        for the ellipse boundary. 2 s.d. are best suited for analysis
        of receptive fields of retinal ganglion cells [1]_. Default is
        2.0

    Returns
    -------
    (mux, muy, r_major, r_minor, rad) : tuple of float, (r,5) np.ndarray
        Parameterized ellipse, where `mux` and `muy` are the centroid,
        `r_major` and `r_minor` are the semi-major and semi-minor axes,
        that is, major and minor radii, and `rad` is the rotation in
        radians. Or stack of `r` parameterized ellipses in an `(r, 5)`
        float array

    Raises
    ------
    IndexError
        If `image` is not two-dimensional.

    ValueError
        If `sigma` is smaller or equal zero.

    Examples
    --------
    The parameterized ellipse offers a compressed representation of the
    ellipse. For example, it is useful for calculating the effective
    diameter of a receptive field.

    >>> p = ellipse(spatial_filter)
    >>> pixel_size = 30  # Conversion to micrometers
    >>> a, b = p[2:4]    # Major and minor axes radii
    >>> rf_diameter = 2 * np.sqrt(a*b) * pixel_size

    To obtain an outline of the ellipse in data coordinates:

    >>> mux, muy, r_major, r_minor, rad = ellipse(spatial_filter)
    >>> num_points = 200
    >>> ls = np.linspace(0, 2*np.pi, num_points)

    >>> rot = [[np.cos(rad), -np.sin(rad)], [np.sin(rad), np.cos(rad)]]
    >>> pos = [r_major * np.cos(ls), r_minor * np.sin(ls)]
    >>> points = np.dot(rot, pos) + [[mux], [muy]]

    References
    ----------
    .. [1] Baden, T., Berens, P., Franke, K., Román Rosón, M., Bethge,
           M., and Euler, T. (2016). The functional diversity of retinal
           ganglion cells in the mouse. Nature, 529(7586), 345–350.
           https://doi.org/10.1038/nature16468
    """
    if sigma <= 0:
        raise ValueError('sigma has to be greater than zero')
    if np.ndim(image) in (1, 3):
        return np.float32([ellipse(i, sigma) for i in image])
    if np.ndim(image) != 2:
        raise IndexError('image is expected to be two-dimensional')

    # Copy and fill NaNs
    s = np.nan_to_num(image, copy=True)

    def gaussian2d(mux, muy, stdx, stdy, rho, amp):
        def fnc(x, y):
            expo = (-(2*(1-rho**2))**-1*((x-mux)**2/stdx**2
                    + (y-muy)**2/stdy**2 - 2*rho*(x-mux)*(y-muy)/(stdx*stdy)))
            with np.errstate(invalid='ignore'):
                mul = (2 * np.pi * stdx * stdy * np.sqrt(1 - rho**2)) ** -1
            return amp * mul * np.exp(expo)
        return fnc

    # Initialize Gaussian parameters around maximum value of s
    max_idx = np.abs(s).argmax()
    max_x, max_y = np.unravel_index(max_idx, s.shape)
    peak = s[max_x, max_y]
    thr = np.exp(-0.5)  # Threshold equivalent to one sigma
    aprx_area = np.count_nonzero(s / peak > thr)
    aprx_r = np.sqrt(aprx_area / np.pi)
    params = np.float32([max_x, max_y, aprx_r, aprx_r, 0, 0])

    # Fit Gaussian
    def lsfun(p):
        return np.reshape(s - gaussian2d(*p)(*np.indices(s.shape)), -1)
    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        mux, muy, stdx, stdy, rho, amp = leastsq(lsfun, params)[0]

    # Create ellipse from Gaussian
    cov = np.array([[stdx**2, rho*stdx*stdy], [rho*stdx*stdy, stdy**2]])
    eigenval, eigenvec = np.linalg.eigh(cov)
    r_major = np.sqrt(eigenval[1]) * sigma  # Largest eigenvalue
    r_minor = np.sqrt(eigenval[0]) * sigma
    with np.errstate(divide='ignore'):
        rad = np.arctan(eigenvec[1, 1] / eigenvec[0, 1])

    # Return parameterized ellipse
    return mux, muy, r_major, r_minor, rad


def contour(image, sigma=1.22, smooth=1.2):
    """
    Create contour outline of an image and return coordinate data points

    Parameters
    ----------
    image : (x,y) or (r,x,y) array_like
        Two-dimensional image or stack of `r` independent
        two-dimensional images, where `x` and `y` are the spatial
        dimensions

    sigma : float, optional
        Standard deviation (s.d.) equivalent to a linearly interpolated
        contour level. A contour level at 1.22 s.d. offers very nice
        visual tiling (for mosaics). Default is 1.22

    smooth : {0.5, 1.2} or float, optional
        Sigma for Gaussian smoothing. This value determines the detail
        of the contour line. The larger the number of pixels, the more
        smoothing is required. As a rule of thumb, 0.5 is suitable for
        subunits and 1.2 for receptive fields. See notes. Default is 1.2

    Returns
    -------
    points : (n,2) or (r,) numpy.ndarray
        Contour outline of `n` data points as xy-coordinate pairs in
        image-pixel coordinates or stack of contours in an array of
        object-type (due to varying number of data points)

    Raises
    ------
    IndexError
        If `image` is not two-dimensional.

    ValueError
        If `sigma` is smaller or equal zero.

    ValueError
        If `smooth` is smaller than zero.

    Notes
    -----
    The image is first nearest-neighbor upsampled and smoothed using a
    `smooth`-sigma Gaussian filter to obtain smooth contour lines. The
    returned contour line is the largest-area island among the positive
    closed contours. The contour is then scaled back down to image-pixel
    resolution.

    When obtaining the contour of a receptive field, a value of 1.2 for
    `smooth` is generally suitable, whereas a value of 0.5 works well
    for the smaller-sized subunits.

    .. note::
        However, suitable values for `smooth` depend on the size of the
        receptive field/subunit in terms of stixels (stimulus/image
        pixels) and may need to be increased for coarser stimuli.

    Examples
    --------
    >>> cntr = contour(spatial_filter)

    Visualize the contour on top of the spatial filter

    >>> from matplotlib import pyplot as plt
    >>> vm = np.nanmax(np.abs(spatial_filter), initial=0) or 1
    >>> pretty = dict(cmap='binary_r', origin='lower', aspect='equal',
    >>>               interpolation='nearest', vmin=-vm, vmax=vm)

    >>> plt.imshow(spatial_filter.T, **pretty)  # Draw image
    >>> plt.plot(*cntr.T, 'r', lw=3)            # Draw contour on top

    Use shapely to perform geometrical analyses

    >>> import shapely
    >>> poly = shapely.Polygon(cntr)
    >>> rf_area = poly.area * pixel_size
    """
    if sigma <= 0:
        raise ValueError('sigma has to be greater than zero')
    if smooth < 0:
        raise ValueError('smooth has to be greater than or equal to zero')
    if np.ndim(image) in (1, 3):
        outl = np.empty(len(image), dtype='object')
        outl[:] = [contour(i, sigma, smooth) for i in image]
        return outl
    if np.ndim(image) != 2:
        raise IndexError('image is expected to be two-dimensional')

    # Constants
    scale = 8 if smooth > 0 else 1
    pad = 15

    # Copy and fill NaNs
    image = np.nan_to_num(image, copy=True)
    image_0 = image.copy()

    # Up-sample with nearest neighbor interpolation
    image = ndimage.zoom(image, scale, order=0, mode='grid-constant',
                         grid_mode=True)

    # Add padding to crop window to avoid open contour at image edges
    image = np.pad(image, pad)

    # Smooth the image
    image = ndimage.gaussian_filter(image, sigma=scale*smooth, mode='nearest')

    # Correct overestimated sigma resulting from smoothing
    if smooth > 0:
        # Ellipse diameter of non-smoothed image
        el = ellipse(image_0, sigma=1)
        diam_nonsmoothed = 2 * np.sqrt(np.prod(el[2:4]))
        # Ellipse diameter of smoothed image
        el = ellipse(image, sigma=1)
        diam_smoothed = 2 * np.sqrt(np.prod(el[2:4])) * 1/scale
        # Rescale sigma with the discrepancy
        diam_min = min(diam_smoothed, 1.75)  # Minimum size
        sigma *= max(diam_nonsmoothed, diam_min) / diam_smoothed

    # Max-scale image to positive unity
    max_idx = np.unravel_index(np.abs(image).argmax(), image.shape)
    peak = image[max_idx]
    image /= peak

    # Contour level is the interpolated s.d.
    contour_level = np.exp(-sigma**2 / 2)
    contours = measure.find_contours(image, contour_level)

    # Select largest-area positive contour (exclude holes and islands)
    contours = [c for c in contours if len(c) >= 4]  # Needed next line
    positive = [c for c in contours if not shapely.LinearRing(c).is_ccw]
    area = lambda c: shapely.Polygon(c).area  # noqa: E731
    points = np.float32(max(positive, key=area, default=np.empty((0, 2))))

    # Interpolate low-resolution contour line (2nd degree B-splines)
    while 0 < len(points) < 100:
        points = measure.subdivide_polygon(points, degree=2)

    # Re-crop and scale back down
    points -= pad  # Remove padding
    points *= 1/scale  # Scale down
    points -= 0.5 * (1 - 1/scale)  # Remove grid offset from scaling
    return points



"""
Matrix factorization classes (:mod:`stnmf.mf`)
==============================================

These classes are decoupled from the concepts of STNMF. The matrix
factorizations follow the nomenclature and descriptors of the
non-negative matrix factorization (NMF) literature. The objective is to
solve

    :math:`\\mathbf{V} \\approx \\mathbf{W}\\mathbf{H} \\,,`

where :math:`V \\in \\mathbb{R}^{n\\times m}` is the input matrix
consisting of :math:`n` variables and :math:`m` observations,
:math:`W \\in \\mathbb{R}^{n\\times r}` are the :math:`r`
recovered features and :math:`H \\in \\mathbb{R}^{r\\times m}` are
the corresponding encodings.

The classes inherit from the class `MF` that decomposes
:math:`V` through iterative updates of :math:`W` and :math:`H` to be
implemented by the sub classes.

.. autosummary::
    :toctree: generated/

    MF
    SemiNMF
    SparseSemiNMF

"""
import numpy as np

from .nnls import afhals
from . import init
from .util import create_rng, trange

__all__ = [
    'SparseSemiNMF',
]


class MF(object):
    """Abstract matrix factorization class"""
    desc = 'Matrix factorization'  #: :class:`str`:Readable analysis name

    def __init__(self, v, r, w0='nnsvdlrc', seed=0, rng=None, dtype='float32'):
        """
        Matrix factorization

        Parameters
        ----------
        v : (n,m) array_like
            Input matrix to decompose

        r : int
            Number of features to recover

        w0 : (n,r) array_like, or {'random', 'nnsvdlrc'}, optional
            If provided, serves as the argument `w0` for the `init`
            function to initialize the features `w`. If None, no
            automatic initialization will be performed, and will have to
            be done manually with `init`. Default is 'nnsvdlrc'

        seed : int, optional
            Random number generator (RNG) seed for reproducibility. If
            None, the RNG is initialized without a seed based on OS
            context. The RNG is only used if initialization is random or
            if using custom callbacks. Default is 0

        rng : numpy.random.Generator, optional
            Random number generator (RNG) instance. If provided, this
            overrides `seed`. Otherwise, the default RNG (twister) will
            be initialized with `seed`. The RNG is only used if
            initialization is random or if using custom callbacks.
            Default is None

        dtype : str or type, optional
            Number format with which to calculate. Default is 'float32'

        Raises
        ------
        ValueError
            If matrix sizes mismatch.

        ValueError
            If `r` is smaller than two.

        See Also
        --------
        stnmf.init : Initialization procedures
        stnmf.util.create_rng : Default RNG instance
        """
        if np.ndim(v) != 2 or min(np.shape(v)) < 2:
            raise ValueError('v is expected to be two-dimensional')
        if r < 2:
            raise ValueError('r has to be at least 2 or larger')

        self.dtype = dtype
        self.r = r

        # Copy V if necessary
        self.v = np.asarray(v, dtype=dtype)
        self.n, self.m = v.shape

        # Allocate memory
        self.h = np.zeros((r, self.m), dtype=dtype)
        """:class:`numpy.ndarray`:Encodings"""
        self.w = np.zeros((self.n, r), dtype=dtype)
        """:class:`numpy.ndarray`:Features"""
        self._res = None

        # Set up random number generator
        if rng is not None:
            self.seed = None
            self.rng = rng
        else:
            self.reseed(seed)

        # Initialize factors
        self.is_initialized = False
        if w0 is not None:
            self.init(w0)

    def init(self, w0='nnsvdlrc', lrc_kwargs=None):
        """
        Initialize features `w`

        Parameters
        ----------
        w0 : (n,r) array_like or {'random', 'nnsvdlrc'}, optional
            Initial features `w`. If array_like, serves as the initial
            `w` directly. If 'random' or 'nnsvdlrd', calls the functions
            from `stnmf.init`. Default is 'nnsvdlrc'

        lrc_kwargs : dict, optional
            Keyword arguments for low-rank correction. Ignored if `w0`
            is not 'nnsvdlrc'. Default is None

        Raises
        ------
        ValueError
            If array sizes mismatch.
        """
        if isinstance(w0, str):
            f = getattr(init, w0)
            if w0 == 'nnsvdlrc':
                method = getattr(self, 'method', self.__class__)
                lrc = dict(method=method, num_iter=1)
                lrc.update(lrc_kwargs or dict())
                arg = lrc  # Parameter lrc for 'nnsvdlrc'
            elif w0 == 'random':
                rng = self.rng
                arg = rng  # Parameter rng for 'random'
            self.w[:] = f(self.v, self.r, arg, self.dtype)
        else:
            if np.shape(w0) != (self.n, self.r):
                raise ValueError('w0 has invalid dimensions')
            np.copyto(self.w, w0)

            # Ensure nonzero columns
            np.nan_to_num(self.w, nan=1e-16, copy=False)
            self.w[:, self.w.sum(axis=0) == 0] = 1e-16

        self.update_h()
        del self.res
        self.is_initialized = True

    def update_w(self):
        """Update features `w`"""
        raise NotImplementedError

    def update_h(self):
        """Update encodings `h` using Moore–Penrose pseudoinverse"""
        np.matmul(np.linalg.pinv(self.w), self.v, out=self.h)

    def residual(self):
        """
        Re-compute reconstruction error

        See Also
        --------
        res : Reconstruction error (residual)
        """
        self._res = np.linalg.norm(self.v - self.w @ self.h, ord='fro')
        return self._res

    @property
    def res(self):
        """
        Reconstruction error (residual) based on Frobenius norm

        :type: :class:`float`
        """
        return self._res or self.residual()

    @res.deleter
    def res(self):
        self._reset_reconstrution()
        self._res = None

    def _reset_reconstrution(self):
        pass

    @np.errstate(divide='ignore', invalid='ignore')
    def factorize(self, num_iter=1000, callback=None, callback_data=None,
                  callback_kwargs=None, disp=True, tqdm_args=None):
        """
        Factorize by iterative updates

        Parameters
        ----------
        num_iter : int, optional
            Number of update iterations. Default is 1000

        callback : function (**kwargs) -> bool, optional
            Callback function to be called after each iteration. See
            notes below for details on the provided function arguments.
            A return value of False will terminate the iteration
            prematurely. Default is None

        callback_data : dict, optional
            Dictionary to store callback data. Depending on the callback
            this dictionary will contain the callback results. Ignored,
            if callback is None. Default is None

        callback_kwargs : dict, optional
            Dictionary serving as additional keyword arguments passed to
            the callback if specified. Ignored, if callback is None.
            Default is None

        disp : bool, optional
            Show progress bar. Equivalent to passing `disabled=True` in
            `tqdm_args`. Default is True

        tqdm_args : dict, optional
            Keyword arguments for `tqdm` progress bar. Ignored if `disp`
            if False. Default is None

        Notes
        -----
        The `callback` function will be supplied with the following
        keyword arguments. Additional arguments can be passed through
        `callback_kwargs`.

        Callback Args
        -------------
        self : stnmf.mf.MF
            Current MF object. Attributes are mutable

        i : int
            Index of the current (completed) iteration. Initialization
            is zero

        itor : tqdm.tqdm
            Iterator used during the factorization. Attributes are
            mutable

        callback_data : dict
            Dictionary to store and preserve callback data. Mutable
        """
        self.num_iter = num_iter  # For later reference
        del self.res

        # Ensure initialization (use defaults)
        if not self.is_initialized:
            self.init()

        # Inlining methods as locals (speed optimization)
        update_w = self.update_w
        update_h = self.update_h

        # Create iterator
        kwargs_t = dict(disable=not disp, desc=self.desc, mininterval=0.5)
        kwargs_t.update(tqdm_args or dict())
        itor = trange(1, num_iter+1, **kwargs_t)
        if itor.pos != 0 and 'leave' not in kwargs_t:
            itor.leave = False

        # Duplicated: If-block outside to avoid 'if' at every iteration
        if callback is None:
            with itor:
                for i in itor:
                    update_w()
                    update_h()
        else:
            # Callback arguments
            if not isinstance(callback_data, dict):
                callback_data = dict()
            kwargs_c = dict(callback_kwargs or {})
            kwargs_c.update(dict(itor=itor, callback_data=callback_data))

            # First callback at zero iterations (after initialization)
            if callback(self, 0, **kwargs_c) is False:
                del self.res
                return

            with itor:
                for i in itor:
                    update_w()
                    update_h()

                    del self.res
                    if callback(self, i, **kwargs_c) is False:
                        break

        del self.res

    def reseed(self, seed=None):
        """
        Re-seed the random number generator (RNG)

        Parameters
        ----------
        seed : int or None, optional
            Random number generator (RNG) seed. If None, the RNG is
            initialized without a seed based on OS context. Default is
            None

        Notes
        -----
        The RNG in this class is only used if initialization is random.
        Calling this method will overwrite the attribute `rng`. It is
        not compatible with using a custom `rng`.

        See Also
        --------
        stnmf.util.create_rng : Create an RNG instance
        """
        self.seed = seed
        self.rng = create_rng(seed)


class SemiNMF(MF):
    """Abstract semi-non-negative matrix factorization (NMF) class"""
    desc = 'Semi-NMF'

    def init(self, *args, **kwargs):
        super().init(*args, **kwargs)

        # Non-negative features
        np.maximum(self.w, 0, out=self.w)


class SparseSemiNMF(SemiNMF):
    """
    Sparse semi-non-negative matrix factorization (NMF)

    Sparse semi-NMF solving

        :math:`\\mathbf{V} \\approx \\mathbf{W}_+\\mathbf{H} \\,,`

    where `w` is non-negative and `v` and `h` are not, using accelerated
    fast HALS algorithms and sparsity constraints.

    See Also
    --------
    stnmf.mf.MF : Abstract matrix factorization class
    """
    desc = 'Sparse semi-NMF'

    def __init__(self, *args, sparsity=1.7, **kwargs):
        """
        Sparse semi-non-negative matrix factorization

        Parameters
        ----------
        sparsity : float, optional
            Regularization weight for sparsity. Default is 1.7

        Keyword Arguments
        -----------------
        kwargs
            See `MF` class

        See Also
        --------
        stnmf.mf.MF : Abstract matrix factorization class
        """
        self.sparsity = sparsity
        super().__init__(*args, **kwargs)

        # Inlining attributes as locals (speed and readability)
        r = self.r
        n = self.n
        m = self.m

        # Compute rho for AF-HALS max iterations assuming non-sparsity
        self.rho_w = 1 + (m*n + n*r) / (m*r + m)
        self.alpha = kwargs.get('alpha', 0.5)
        self.eps = kwargs.get('eps', 0.1)

        # Pre-allocate memory for in-place operations
        self._vht = np.empty((n, r), dtype=self.dtype)
        self._hht = np.empty((r, r), dtype=self.dtype)

    def init(self, *args, lrc_kwargs=None, **kwargs):
        lrc_kwargs = lrc_kwargs or dict()
        lrc_kwargs.setdefault('sparsity', self.sparsity)
        super().init(*args, lrc_kwargs=lrc_kwargs, **kwargs)

    def update_h(self):
        """Update encodings `h` using Moore–Penrose pseudoinverse"""
        super().update_h()

        # Normalize rows
        h = self.h
        h /= np.linalg.norm(h, axis=1, keepdims=True)

    def update_w(self):
        """Update features `w` using AF-HALS"""

        # Inlining attributes as locals (speed optimization)
        h = self.h
        ht = h.T
        vht = self._vht
        hht = self._hht

        np.matmul(self.v, ht, out=vht)
        np.matmul(h, ht, out=hht)
        afhals(vht, hht, self.w, rho=self.rho_w, alpha=self.alpha,
               eps=self.eps, lmbda=self.sparsity, inplace=True)



"""
Example callback to implement a custom stopping criterion
=========================================================
"""
import numpy as np

__all__ = [
    'callback',
]


def callback(self, i, itor, callback_data, interval=100, delta=1e-6):
    """
    Callback function to implement a custom stopping criterion based on
    relative reconstruction improvement

    Parameters
    ----------
    interval : int, optional
        Evaluate stopping criterion every `interval` iterations

    delta : float, optional
        Fraction of first-iteration improvement. When the difference to
        the previous reconstruction error falls below, terminate the
        factorization. Default is 1e-6

    Notes
    -----
    This callback function is not specific to :class:`stnmf.STNMF` and
    can be used in conjunction with other :mod:`snmt.mf` classes.

    .. warning:: The reconstruction residual is not a suitable metric
                 reconstruction quality in the scope of
                 :class:`stnmf.STNMF`, due to the sparsity involved!

    Examples
    --------
    >>> from stnmf import STNMF
    >>> from stnmf.callbacks import stoppping_criterion
    >>> stnmf = STNMF(ste, callback=stoppping_criterion,
    ...               callback_kwargs=dict(interval=100))
    """
    if i == 0:
        # Store the initial reconstruction residual for comparison
        callback_data['err_0'] = self.res
        callback_data['err_prev'] = np.inf

    elif i == 1:
        # Define threshold of improvement based on first iteration
        callback_data['thr'] = (callback_data['err_0'] - self.res) * delta

    elif i % interval == 0:
        # Termination criterion
        if callback_data['err_prev'] - self.res < callback_data['thr']:
            itor.close()
            return False

        # Otherwise continue and update the previous error
        callback_data['err_prev'] = self.res



"""
Utilities (:mod:`stnmf.util`)
=============================

Utility functions for internal use

.. autosummary::
    :toctree: generated/

    create_rng
    tqdm

"""
import numpy as np
from tqdm import tqdm as std_tqdm

__all__ = [
    'create_rng',
]


def create_rng(seed=None):
    """
    Obtain a new seeded instance of a pseudorandom number generator
    compatible across Python, MATLAB, and R

    Mersenne Twister (MT19937) [1]_.

    Parameters
    ----------
    seed : int, optional
        Initial seed. If None, the RNG will be initialized without a
        seed based on OS context. Default is None

    Returns
    -------
    rng : numpy.random.Generator
        The seeded random number generator

    Notes
    -----
    For exact compatibility with MATLAB, sequences should be sampled in
    column-order, e.g. two-dimensional arrays should be transposed after
    sampling.

    .. warning:: Only basic sampling methods like `random` yield
                 identical sequences of values as in MATLAB. More
                 complex functions like `choice` are not compatible with
                 the twister RNG of MATLAB!

    References
    ----------
    .. [1] Matsumoto, & Nishimura. (1998). Mersenne twister. ACM
           Transactions on Modeling and Computer Simulation, 8(1), 3–30.
           https://doi.org/10.1145/272991.272995
    """
    mt = np.random.MT19937()
    mt.state = np.random.RandomState(seed).get_state()
    rng = np.random.Generator(mt)
    return rng


def tqdm(*args, **kwargs):
    """
    Wrapper function for tqdm exposing the attribute `auto` to switch
    between `tqdm.auto.tqdm` and `tqdm.std.tqdm`.

    Attributes
    ----------
    auto : bool
        If True, use `tqdm.auto.tqdm`. If False resort back to
        `tqdm.std.tqdm`. Default is True

    Notes
    -----
    While `tqdm.auto.tqdm` can distinguish Python from IPython for
    displaying respective progress bars appropriately, jupyter notebooks
    are indistinguishable from the jupyter console. To avoid issues, the
    use of `tqdm` within this package can be adjusted. This here
    function wraps `tqdm` and allows to override its use and to manually
    disable auto detection if desired by setting `stnmf.util.tqdm.auto`
    to False.

    Examples
    --------
    Override the setting and revert to the standard `tqdm` and `trange`.

    >>> import stnmf.util
    >>> stnmf.util.tqdm.auto = False
    """
    if getattr(tqdm, 'auto', True):
        from tqdm.auto import tqdm as auto_tqdm
        return auto_tqdm(*args, **kwargs)
    else:
        return std_tqdm(*args, **kwargs)


def trange(*args, **kwargs):
    return tqdm(range(*args), **kwargs)



"""
Consensus analysis
==================
"""
import numpy as np
from scipy.cluster.hierarchy import linkage, cophenet

from ..util import tqdm

__all__ = [
    'callback',
    'cpcc',
    'affiliation',
]


def callback(self, i, itor, callback_data, sparsities, num_rep=30,
             max_elem=25000):
    """
    Callback function to run consensus analysis

    The results are provided in a :mod:`numpy.ndarray` as cophenetic
    correlation coefficients stored in the key 'cpcc' in the dictionary
    `callback_data`.

    Parameters
    ----------
    sparsities : (n,) array_like
        List of `n` regularization parameter values to probe

    num_rep : int, optional
        Number of repetitions for each parameter to check consensus
        over. Default is 30

    max_elem : int, optional
        Maximum number of columns to consider from the encodings matrix.
        Reducing this value is beneficial if the encodings matrix
        has many columns, i.e. if the STNMF weight factor matrix has
        many spikes. Higher values may require a lot of memory. Default
        is 25000

    Notes
    -----
    Selecting `max_elem` elements is done by sampling without
    replacement. The indices are selected at initialization and are
    retained throughout the analysis to always compare the same
    elements.

    .. warning:: Even with identical random number generator and seed,
                 the implementation of choice-without-replacement
                 differs from MATLAB. The consensus analysis with
                 `max_elem < m` is thus not compatible with MATLAB!

    .. note:: Implemented for finding the suitable sparsity
              regularization parameter for subunit recovery with
              :class:`stnmf.STNMF` only. No other parameter type or
              :mod:`stnmf.mf` class is supported.

    For more details on the implementation, visit the source code.

    Examples
    --------
    The code below runs the consensus analysis on a given parameter
    range.

    >>> from stnmf import STNMF
    >>> from stnmf.callbacks import consensus
    >>> results = dict()
    >>> STNMF(ste, callback=consensus, callback_data=results,
    ...       callback_kwargs=dict(sparsities=[0, 0.5, 1.0, 1.5, 2]))
    >>> print(results['cpcc'])
    """
    if i == 0:
        # Initialize before the first iteration
        callback_data['_obj'] = Analysis(self, itor, num_rep, max_elem,
                                         sparsities, callback_data)
    elif i % self.num_iter == 0:
        # Complete repetitions of current parameter
        callback_data['_obj'].advance(i)
    else:
        # Pass through and update progress bar
        callback_data['_obj'].pass_through()


def cpcc(affil):
    """Calculate the cophenetic correlation of an affiliation matrix"""
    affil = np.asarray(affil)
    num_rep, m = affil.shape

    # Create consensus matrix
    cons = np.ones(m * (m-1) // 2, dtype='float32')  # Precision vs. memory
    idx = 0
    for k in range(1, m):
        part = slice(idx, idx + m-k)
        np.equal(affil[:, k:], affil[:, k-1, None]).sum(axis=0, out=cons[part])
        idx += m-k
    cons /= num_rep

    # Get linkage
    cons = np.subtract(1, cons, out=cons)
    z = linkage(cons, method='average')

    # Calculate cophenetic correlation
    with np.errstate(invalid='ignore'):
        coph = cophenet(z, cons)[0]

    return coph


def affiliation(h, elem=slice(None), localized=[], out=None):
    """Create affiliation matrix from encodings"""
    h = np.asarray(h)
    if out is None:
        m = h.shape[1] if isinstance(elem, slice) else len(elem)
        out = np.empty(m, dtype=h.dtype)
    out[:] = np.abs(h[:, elem]).argmax(axis=0)  # Casting does not allow 'out'
    non_loc = np.isin(out, localized, invert=True)
    out[non_loc] = np.nan
    return out


class Analysis:
    """Class for consensus analysis callback"""
    def __init__(self, mf, itor, num_rep, max_elem, sparsities, data):
        assert mf.__class__.__name__ == 'STNMF', 'Supports STNMF only'

        self.mf = mf
        self.num_rep = num_rep
        self.sparsities = sparsities
        self.data = data

        # Allocate arrays
        max_elem = np.clip(max_elem, 0, mf.m) or mf.m
        self.coph = np.full(len(sparsities), np.nan, mf.dtype)
        self.affil = np.full((num_rep, max_elem), np.nan, mf.dtype)
        if mf.m > max_elem:
            # This is not compatible with MATLAB's RNG implementation!
            self.elem = mf.rng.choice(mf.m, max_elem, replace=False)
        else:
            self.elem = slice(None)

        # Adjust iterator and replace iterable
        itor.total = len(sparsities) * num_rep * mf.num_iter
        itor.iterable = range(1, itor.total+1)

        # Create and adjust progress bars
        disabled = itor.disable
        leave = itor.pos == 0
        mininterval = itor.mininterval if not disabled else 0.5
        desc = itor.desc if not disabled else mf.desc
        lpad = max(map(len, [desc, 'Consensus analysis']))
        itor.leave = False
        itor.close()
        self.it_p = tqdm(sparsities, desc='Consensus analysis'.rjust(lpad),
                         unit='parameter', leave=leave, disable=disabled,
                         miniters=1, postfix=dict(sparsity=sparsities[0]))
        self.it_r = tqdm(total=num_rep, desc='Repetition'.rjust(lpad),
                         unit='rep', leave=False, disable=disabled,
                         miniters=1, postfix=dict(seed=0))
        self.it_i = tqdm(total=mf.num_iter, desc=desc.rjust(lpad), leave=False,
                         mininterval=mininterval, disable=disabled)

        # Update callback data for user output and self-reference
        data['_obj'] = self
        data['cpcc'] = self.coph

        # Saved for user if requrested
        if 'affiliation_matrix' in data:
            data['affiliation_matrix'] = self.affil
        if 'selected_indices' in data:
            data['selected_indices'] = self.elem

        # Initialize first run of STNMF
        mf.sparsity = sparsities[0]
        mf.reseed(0)
        mf.init('random')

    def advance(self, i):
        """Advance the consensus analysis after completed repetition"""
        # Inlining attributes as locals (speed and readability)
        mf = self.mf
        it_p = self.it_p
        it_r = self.it_r
        it_i = self.it_i

        # Update progress bars
        it_r.update()
        it_i.reset()
        it_i.clear()
        it_p.n = min(it_p.total, round(it_p.n + 1/it_r.total, 2))
        if it_p.n == round(it_p.n):
            it_p.n = int(it_p.n)
        it_p.update(0)

        # Find current iteration
        total_rep = int((i-1) // mf.num_iter)
        cur_param, cur_rep = divmod(total_rep, self.num_rep)

        # Store affiliations of this repetition
        if mf.num_subunits:  # Only consider localized subunits
            affiliation(mf.h, self.elem, mf.localized, out=self.affil[cur_rep])

        # Calculate consensus among repetitions for this parameter
        if cur_rep == self.num_rep-1:
            if not it_r.disable:
                num_spaces = getattr(it_r, 'ncol', 0)
                it_r.display('Computing consensus...'.ljust(num_spaces))
            self.coph[cur_param] = cpcc(self.affil)

            # Go to next parameter
            it_p.n = min(it_p.total, cur_param+1)
            it_p.update(0)

            # Terminate
            if it_p.n >= it_p.total:
                self.terminate()
                return

            # Reset to next parameter
            mf.seed = 0
            mf.sparsity = self.sparsities[it_p.n]
            it_p.set_postfix(sparsity=mf.sparsity)
            it_r.reset()
            self.affil.fill(np.nan)
        else:
            # Advance to next repetition
            mf.seed += 1

        # Initialize for next run of STNMF
        it_r.set_postfix(seed=mf.seed)
        mf.reseed(mf.seed)
        mf.init('random')

    def pass_through(self):
        """Update progress bar only"""
        self.it_i.update()

    def terminate(self):
        """Make sure the progress bars close on termination"""
        self.it_p.set_postfix(None)
        self.it_p.clear()

        # Close all progress bars
        self.it_i.close()
        self.it_r.close()
        self.it_p.close()

        # Clear temporary values
        try:
            del self.data['_obj']
        except KeyError:
            pass

    def __del__(self):
        self.terminate()



"""
Perturbations
=============
"""
import numpy as np

from ..util import create_rng

__all__ = [
    'apply',
    'callback',
]


def callback(self, i, itor, callback_data, interval=100):
    """
    Callback function to apply different perturbations to the modules at
    certain intervals

    Notes
    -----
    Implemented for subunit recovery with :class:`stnmf.STNMF` only. No
    other :mod:`stnmf.mf` class is supported.

    Different perturbations are chosen based on random sampling. This
    should be largely compatible with the random choice implementation
    of MATLAB.

    Examples
    --------
    The code below applies a random perturbation of the modules every
    `interval` iterations.

    >>> from stnmf import STNMF
    >>> from stnmf.callbacks import perturbation
    >>> stnmf = STNMF(ste, callback=perturbation,
    ...               callback_kwargs=dict(interval=100))
    """
    if i == 0:
        # Allocate arrays for callback
        assert self.__class__.__name__ == 'STNMF', 'Supports STNMF only'
        callback_data['res_prev'] = np.inf
        callback_data['modules_prev'] = np.zeros_like(self.modules)
        callback_data['weights_prev'] = np.zeros_like(self.weights)
    elif i % interval == 0:
        res_prev = callback_data['res_prev']
        modules_prev = callback_data['modules_prev']
        weights_prev = callback_data['weights_prev']
        disp = not itor.disable

        printmsg = f'{i:7}it: {self.res:0.4f}'

        if self.res > res_prev:
            printmsg += f' > {res_prev:0.2f} -> Revert.'
            self.modules = modules_prev
            self.weights = weights_prev
        else:
            printmsg += ''.ljust(21)
            np.copyto(modules_prev, self.modules)
            np.copyto(weights_prev, self.weights)
            callback_data['res_prev'] = self.res

        if i < self.num_iter:
            ret = apply(self.modules, self.localized, disp=disp, rng=self.rng,
                        inplace=not self.masked)
            if disp:
                msg = ret[2]
                printmsg += f' {msg}'
                ret = ret[0]
            if self.masked:
                self.modules = ret

        if disp:
            itor.write(printmsg, file=itor.fp)


def apply(modules, localized=[], disp=True, inplace=False, rng=None):
    """
    Apply randomly chosen perturbation to spatial modules

    Parameters
    ----------
    modules : (r,x,y) array_like
        Modules to perturb with spatial dimensions `x` and `y`

    localized : (l,) array_like, optional
        Indices of localized modules, where `l` is the number of
        localized modules. If empty, all modules are considered
        non-localized. Default is []

    disp : bool, optional
        Return additional information about the perturbation performed.
        Default is True

    inplace : bool, optional
        If True, operate on `modules` in-place. Default is False

    rng : numpy.random.Generator or int or None, optional
        Random number generator (RNG) instance. If integer, it will be
        the seed for initializing a new RNG. If None, the RNG will be
        initialized without a seed based on OS context. Default is None

    Returns
    -------
    modules : (r,x,y) numpy.ndarray
        Perturbed modules

    p : int
        Perturbation type that was applied. -1 if none was applicable.
        This value is only returned if `disp` is True.

    msg : str or None
        Perturbation description. None if no perturbation was performed.
        This value is only returned if `disp` is True.
    """
    modules = np.array(modules, copy=not inplace)
    r = modules.shape[0]

    # Initialize RNG
    if rng is None or isinstance(rng, int):
        rng = create_rng(seed=rng)

    # Random choice implementation to match MATLAB's behavior
    def choice(arr):
        """
        Equivalent to MATLAB's
          `arr(randperm(numel(arr), 1))`
        and
          `datasample(arr, 1)` (stats toolbox)
        """
        return rng.choice(arr, p=np.repeat(1/len(arr), len(arr)))

    # Determine localized and non-localized modules
    nonlocalized = np.nonzero(np.isin(range(r), localized))[0]

    # Determine which perturbations are applicable
    pert_candidates = []
    if localized.size:
        pert_candidates.append(reseed)
        module_local = modules[choice(localized)]
    if nonlocalized.size:
        pert_candidates.append(0)
        module_nonlocal = modules[choice(nonlocalized)]
    if localized.size and nonlocalized.size:
        pert_candidates += [duplicate, splitvert, splithorz]

    msg, p = None, -1
    if pert_candidates:
        p = choice(pert_candidates)

        if p:
            msg = p(module_local, module_nonlocal, rng=rng)
        else:
            # Advanced indexing does not allow in-place modification
            m = modules[nonlocalized]
            msg = reseed(m, rng=rng)
            msg += 's (all non-localized)'
            modules[nonlocalized], m = m, None

    return modules, p, msg if disp else modules


def reseed(modules, *args, rng):
    """
    Re-seed modules
    """
    msg = 'Re-seed module'
    modules[:] = rng.random(modules.shape[::-1], dtype=modules.dtype).T
    return msg


def duplicate(module_src, module_dst, rng):
    """
    Duplicate a module
    """
    msg = 'Duplicate a module and add noise to both copies'
    i_max = np.unravel_index(np.nanargmax(np.abs(module_src)),
                             module_src.shape)
    m_max = module_src[i_max]
    module_dst[:] = module_src
    module_src[:] += rng.random(module_src.shape[::-1]).T * m_max
    module_dst[:] += rng.random(module_src.shape[::-1]).T * m_max
    return msg


def splitvert(module_src, module_dst, rng):
    """
    Split a module horizontally
    """
    msg = 'Split one localized module into halves vertically'
    i_max = np.unravel_index(np.nanargmax(np.abs(module_src)),
                             module_src.shape)[0]
    module_dst[i_max:] = module_src[i_max:]
    module_dst[:i_max] = 0
    module_src[i_max:] = 0
    return msg


def splithorz(module_src, module_dst, rng):
    """
    Split a module horizontally
    """
    msg = 'Split one localized module into halves horizontally'
    i_max = np.unravel_index(np.nanargmax(np.abs(module_src)),
                             module_src.shape)[1]
    module_dst[:, i_max:] = module_src[:, i_max:]
    module_dst[:, :i_max] = 0
    module_src[:, i_max:] = 0
    return msg



"""
Initialization of matrix factorization (:mod:`stnmf.init`)
==========================================================

Collection of functions that offer strategies for initialization of
matrix factorization.

Initialization largely contributes to convergence time and the solution
that is found [1]_ [2]_.

.. autosummary::
    :toctree: generated/

    nnsvdlrc
    random

Notes
-----
The functions provided can be called directly by the matrix
factorization classes by passing their name as the argument `w0`.

Examples
--------
The code below is equivalent.

>>> w = stnmf.init.nnsvdlrc(v, r=20)
>>> stnmf.mf.SparseSemiNMF(v, r=20, w0=w)

>>> stnmf.mf.SparseSemiNMF(v, r=20, w0='nnsvdlrc')

References
----------
.. [1] Cichocki, A., Zdunek, R., Phan, A. H., & Amari, S. (2009).
       Nonnegative Matrix and Tensor Factorizations (1st ed.).
       Chichester, UK: John Wiley & Sons, Ltd.
       https://doi.org/10.1002/9780470747278

.. [2] Atif, S. M., Qazi, S., & Gillis, N. (2019). Improved SVD-based
       initialization for nonnegative matrix factorization using
       low-rank correction. Pattern Recognition Letters, 122, 53–59.
       https://doi.org/10.1016/J.PATREC.2019.02.018
"""
import numpy as np
from scipy.sparse.linalg import svds

from .util import create_rng

__all__ = [
    'nnsvdlrc',
    'random',
]


def random(v, r, rng=None, dtype='float32'):
    """
    Initialize feature matrix by random sampling

    Parameters
    ----------
    v : (n,m) array_like
        Input matrix

    r : int
        Number of components

    rng : numpy.random.Generator or int or None, optional
        Random number generator (RNG) instance. If integer, it will be
        the seed for initializing a new RNG. If None, the RNG will be
        initialized without a seed based on OS context. Default is None

    dtype : str or type, optional
        Number format with which to calculate. Default is 'float32'

    Returns
    -------
    w : (n,r) numpy.ndarray
        Initialized `w`

    See Also
    --------
    stnmf.util.create_rng : Create an RNG instance
    """
    if rng is None or isinstance(rng, int):
        rng = create_rng(seed=rng)

    n, m = np.shape(v)
    w = rng.random((r, n), dtype).T  # Transpose: MATLAB compatibility
    return w


def nnsvdlrc(v, r, lrc=True, dtype='float32'):
    """
    Initialize feature matrix with a variation of non-negative singular
    value decomposition with low-rank correction (NNSVD-LRC)

    For details on NNSVD-LRC [1]_ and its modifications to the algorithm
    see notes below.

    Parameters
    ----------
    v : (n,m) array_like
        Input matrix

    r : int
        Number of components

    lrc : dict or bool, optional
        Perform subsequent low-rank correction as matrix factorization
        on the low-rank approximation `v` to improve the initial
        features `w`. This is considered computationally cheaper than
        factorization on the full `v`. If `lrc` is True, use default
        parameters for creating a `MF` object, with one iteration and
        method 'SparseSemiNMF'. If dict, it serves as keyword arguments
        for initialization of the `MF` object, incl. `num_iter`. See
        `stnmf.MF` for details.

    dtype : str or type, optional
        Number format with which to calculate. Default is 'float32'

    Returns
    -------
    w : (n,r) numpy.ndarray
        Initialized `w`

    Notes
    -----
    Since the input matrix `v` is not non-negative in a semi-NMF/MF
    problem, the first rank-one factor of the SVD is not exclusively
    positive, but may contain positive and negative values. Therefore,
    step 2 in the algorithm of [1]_ is skipped here and step 3 is
    performed on all `r` rank-one factors.
    Additionally, unlike in [1]_, only `w` is initialized, as `h` is to
    be inferred subsequently.

    References
    ----------
    .. [1] Atif, S. M., Qazi, S., & Gillis, N. (2019). Improved
           SVD-based initialization for nonnegative matrix factorization
           using low-rank correction. Pattern Recognition Letters, 122,
           53–59. https://doi.org/10.1016/J.PATREC.2019.02.018
    """
    v = np.asarray(v, dtype=dtype)
    n, m = v.shape

    # Initialize the truncated SVD deterministically by all ones
    init = np.ones(np.min(v.shape), dtype=dtype)

    # Compute truncated SVD with p components (ceil r/2)
    p = int((r+1)/2)
    u, s, vt = svds(v, k=p, v0=init, maxiter=np.iinfo(np.int32).max,
                    random_state=0)

    # svds does not guarantee any ordering
    order = np.argsort(s)[::-1]
    u = u[:, order]
    s = s[order]

    # Integrate the singular values into the components
    s = np.sqrt(np.diag(s))
    w_p = u @ s

    # Flip sign for reproducible component order under sign ambiguity
    max_i = np.abs(w_p).argmax(axis=0)
    sign = np.copysign(1, w_p[max_i, range(p)])
    w_p *= sign[None, ...]

    # Create empty array for W
    w = np.zeros((n, r), dtype=dtype)

    # Add each factor twice: Once positive, once negative inverted
    np.maximum(w_p, 0, out=w[:, ::2])
    np.maximum(-w_p[:, :int(r/2)], 0, out=w[:, 1::2])

    # Ensure nonzero columns/rows
    np.nan_to_num(w, nan=1e-16, copy=False)
    w[:, w.sum(axis=0) == 0] = 1e-16

    # Perform low-rank correction on the low-rank approximation of V
    if lrc is not False:
        h_p = s @ vt[order] * sign[..., None]
        v_p = w_p @ h_p  # Low-rank approximation of V

        # Obtain arguments for factorization
        mf_args = lrc if isinstance(lrc, dict) else dict()
        method = mf_args.pop('method', 'SparseSemiNMF')
        lowrank_iter = mf_args.pop('num_iter', 1)

        # Obtain matrix factorization class
        if isinstance(method, str):
            from . import mf
            mfclass = getattr(mf, method)
        else:
            mfclass = method

        # Run factorization iterations
        fact = mfclass(v_p, r, w0=w, dtype=dtype, **mf_args)
        fact.factorize(num_iter=lowrank_iter, disp=False)
        w = fact.w

    return w



"""
Spike-triggered non-negative matrix factorization
=================================================
"""
import numpy as np
import shapely
import warnings

from . import mf
from . import spatial
from .plot import all as plot_all

__all__ = [
    'STNMF',
]


class STNMF(mf.MF):
    """
    Spike-triggered non-negative matrix factorization (STNMF)

    An object of the class exposes the following attributes and methods.
    For details, see the documentation of the individual items below.

    Spatial modules and weights
    ---------------------------
    .. autosummary::
      modules       -- Recovered modules
      weights       -- Recovered weights
      subunits      -- Localized subunits (norm. subset of modules)
      ratios        -- Scalar subunit weights (norm. weight)
      num_subunits  -- Number of localized subunits

    Subunit attributes
    ------------------
    .. autosummary::
      outlines      -- Contour outlines
      diameters     -- Diameters in micrometers (μm)
      distances     -- Distances from STA center in micrometers (μm)
      polarities    -- Polarities (positive or negative)
      moransi       -- Moran's I autocorrelation
      localized     -- Indices of localized modules

    Object methods
    --------------
    .. autosummary::
      plot          -- Visualize the STNMF results in a figure
      factorize     -- Inherited function from `stnmf.mf.MF`
      flipsign      -- Flip sign of modules to make weights positive
      sort          -- Sort modules by autocorrelation and weight

    Parameter attributes
    --------------------
    .. autosummary::
      pixel_size    -- Pixel size in micrometers (μm)
      moransi_threshold -- Threshold to distinguish localized modules

    Further attributes
    ------------------
    .. autosummary::
      sta           -- Spike-triggered average (STA), mean of STE
      sta_outline   -- Contour outline of the STA
      sta_diameter  -- Diameter of the STA in micrometers (μm)
      coverage      -- Spatial coverage factor of subunits on STA
      modules_info  -- Collection of the properties for all modules
      res           -- Reconstruction residual

    See Also
    --------
    stnmf.mf : Matrix factorization classes

    Examples
    --------
    >>> from stnmf import STNMF
    >>> stnmf = STNMF(ste)

    The created STNMF object exposes further attributes and functions;
    for example, to visualize the resulting decomposition.

    >>> stnmf.plot()

    Notes
    -----
    Find more information in the documentation of the `__init__` method
    below.
    """
    method_default = 'SparseSemiNMF'

    def __new__(cls, *args, **kwargs):
        """Dynamically create appropriate class from provided method"""
        method = kwargs.get('method', STNMF.method_default)
        if isinstance(method, str):
            mfclass = getattr(mf, method)
        else:
            mfclass = method
        kwargs['method'] = mfclass

        clss = type('STNMF', (mfclass,), dict(STNMF.__dict__))
        clss.factorize.__doc__ = super().factorize.__doc__

        obj = mfclass.__new__(clss)
        obj.__init__(*args, **kwargs)
        return obj

    def __init__(self, ste, r=20, init='nnsvdlrc', flipsign=True,
                 sort=True, factorize=True, *args, **kwargs):
        """
        Spike-triggered non-negative matrix factorization using AF-HALS

        Parameters
        ----------
        ste : (x,y,m) array_like
            Spike-triggered stimulus ensemble with `x*y` pixels and `m`
            spikes

        r : int, optional
            Number of modules to recover. Default is 20

        init : (r,x,y) array_like or {'random', 'nnsvdlrc'}, optional
            Initial modules. If array_like, serves as the initial
            modules directly. If 'random' or 'nnsvdlrc', calls the
            functions from `stnmf.init`. If None, no automatic
            initialization will be performed, and has to be done
            manually by calling `init`. Default is 'nnsvdlrc'

        flipsign : bool, optional
            Flip the sign of the modules to make their mean weights
            positive. Default is True

        sort : bool, optional
            Sort modules by mean weights. Default is True

        factorize : bool, optional
            Factorize on initialization. Default is True

        Keyword Arguments
        -----------------
        method : str or class, optional
            Matrix factorization class. Default is 'SparseSemiNMF'

        sparsity : float, optional
            Sparsity regularization weight if `method` is
            'SparseSemiNMF'; otherwise invalid parameter. Default is 1.7

        seed : int, optional
            Random seed for reproducibility. Default is 0

        dtype : str, type or numpy.dtype, optional
            Number format with which to calculate. Default is 'float32'

        pixel_size : float, optional
            Size of a stixel (stimulus pixel) in micrometers (μm) for
            calculation of spatial diameter and distance. If not set,
            these calculations yield zero. Default is 0

        moransi_threshold : float, optional
            Moran's I (autocorrelation) by which to distinguish
            localized from non-localized modules. Default is 0.25

        Factorization Args
        ------------------
        kwargs
            See :func:`STNMF.factorize`

        Raises
        ------
        ValueError
            If `ste` has incorrect dimensions.

        Notes
        -----
        If `factorize` is True (default), the decomposition into
        subunits is performed immediately with the provided
        `Factorization Args`. If set to False, the factors are
        initialized. The factorization is then started manually by
        calling `factorize` with applicable arguments. In that case, any
        `Factorization Args` meant for the factorization that are passed
        to the initialization are invalid and will raise an error.

        The `ste` may contain masked or NaN values to speed up the
        decomposition.

        .. warning::
           To obtain appropriate `diameters` and `distances`, it is
           essential to provide a value for `pixel_size` or to assign
           the attribute later.

        See Also
        --------
        stnmf.mf : Matrix factorization classes with additional keyword
            arguments
        factorize : For more keyword arguments
        """
        if np.ndim(ste) != 3:
            raise ValueError('ste is expected to have three dimensions')

        # Format input
        self.x, self.y = np.shape(ste)[:2]

        # Handle mask and NaN
        if np.ma.is_masked(ste):
            mask = ~ste.mask.sum(axis=-1, dtype='bool')
        else:
            mask = np.isfinite(np.sum(ste, axis=-1))
        if not np.all(mask):
            v = np.asarray(ste)[mask]
            self.mask = (Ellipsis,) + np.nonzero(mask)
            self.masked = True
        else:
            v = np.reshape(ste, (self.x*self.y, -1))
            self.mask = slice(None)
            self.masked = False

        self._sta = np.mean(ste, axis=-1)
        self._sta.setflags(write=False)
        self.pixel_size = kwargs.pop('pixel_size', 0)  # Default none
        self.moransi_threshold = kwargs.pop('moransi_threshold', 0.25)
        self.autoflip = flipsign
        self.autosort = sort

        # Format initialization
        if init is None or isinstance(init, str):
            kwargs['w0'] = init
        elif self.masked:
            kwargs['w0'] = np.asarray(init)[self.mask].T
        else:
            kwargs['w0'] = np.reshape(init, (r, self.x*self.y)).T

        # Pick apart kwargs for initialization and factorization
        if factorize:
            # If supplying these but factorize is False, throw an error!
            ky = ['num_iter', 'callback', 'callback_data', 'callback_kwargs',
                  'disp', 'tqdm_args']
            kwargs_fac = {k: kwargs.pop(k) for k in ky if k in kwargs}

        # Initialize matrix factorization
        self.method = kwargs.pop('method', STNMF.method_default)
        super(self.__class__, self).__init__(v=v, r=r, *args, **kwargs)

        # Create modules and weights
        self._weights = self.h
        if self.masked:
            # Modules are a non-modifiable copy
            dim = (self.r, self.x, self.y)
            self._modules = np.full_like(self.w, np.nan, shape=dim)
            self._modules[self.mask] = self.w.T
            self._modules.setflags(write=False)
        else:
            # Modules are a modifiable view
            self._modules = self.w.T.reshape(self.r, self.x, self.y)

        # Factorize immediately
        if factorize:
            self.factorize(**kwargs_fac)

    def _reset_localized(self):
        """Reset cached attributes based on localized subunits"""
        self._moransi = None
        self._outlines = None
        self._diameters = None
        self._distances = None
        self._polarities = None

    def _reset_sized(self):
        """Reset cached attributes based on pixel size"""
        self._outlines = None
        self._diameters = None
        self._distances = None
        self._sta_outline = None
        self._sta_diameter = None

    def _reset_reconstrution(self):
        """Reset cached attributes based on reconstruction"""
        self._reset_localized()

    def factorize(self, *args, **kwargs):
        super(self.__class__, self).factorize(*args, **kwargs)
        if self.autoflip:
            self.flipsign()
        if self.autosort:
            self.sort()

    def plot(self, full=True, colors='hsv', cmap=None, **kwargs):
        """
        Visualize STNMF results in a figure

        Parameters
        ----------
        full : bool, optional
            Display the full figure. If False, omit weights and
            autocorrelation plots (bottom row). Default is True

        colors : color, iterable, matplotlib.colors.Colormap, optional
            Single color or color sequence to differentiate localized
            modules. If single color, the same color is used on the
            plots of all localized subunits. If iterable or colormap,
            colors are iterated over plots of localized subunits.
            Default is 'hsv'

        cmap : str or matplotlib.colors.Colormap, optional
            Override colormaps for both STA and modules. By default the
            modules and STA plots have two different colormaps. If
            specified, the colormap is used for all subplots.
            Recommended are 'rwb' for convenient view (red-white-black
            with masked values lightseagreen) and 'filter' for
            conventional spatial contrast filter colors
            (black-gray-white with masked values magenta). If None,
            modules are visualized with 'rwb' and STA with 'filter'.
            Default is None

        Keyword Arguments
        -----------------
        kwargs
            Further arguments may be passed to the plotting routines

        Returns
        -------
        fig : matplotlib.figure.Figure
            Created figure

        Warns
        -----
        PixelSizeWarning
            If no scale bar will be drawn due to missing pixel size

        See Also
        --------
        stnmf.plot.all : Create figure combining all STNMF results
        """
        self._pixel_size_warning()
        kwargs.update(dict(
            w=self.weights if full else None,
            autocorr=self.moransi,
            pixel_size=self.pixel_size or None,
            colors=colors,
            cmap=cmap,
            moransi_threshold=self.moransi_threshold,
        ))
        fig = plot_all(self.sta, self.modules, self.sta_outline, self.outlines,
                       **kwargs)
        return fig

    def flipsign(self):
        """
        Flip sign of modules for their average weights to be positive
        """
        idx = self.weights.mean(axis=1) < 0
        self.weights[idx] = -self.weights[idx]
        self.w[:, idx] = -self.w[:, idx]  # Use W for views and copies

    def sort(self):
        """
        Sort modules by their absolute average weight, localized first
        """
        idx = np.argsort(np.abs(self.weights.mean(axis=1)))[::-1]
        idx_loc = np.isin(idx, self.localized)
        idx0 = idx[idx_loc]
        idx1 = idx[~idx_loc]
        idx = np.append(idx0, idx1)

        self.modules = self.modules[idx]
        self.weights = self.weights[idx]
        self._reset_localized()

    """
    Spatial modules and weights
    ---------------------------
    """

    @property
    def modules(self):
        """
        Recovered spatial modules

        :type: :class:`numpy.ndarray`

        Notes
        -----
        This array is a reshaped view of the matrix factorization
        features `stnmf.mf.MF.w`.

        However, if the STE contains NaNs or is otherwise masked (i.e.
        `masked` is True), this array is an immutable copy that
        can only be modified by direct assignment.

        See Also
        --------
        subunits : Localized subunits only
        stnmf.mf.MF.w : Matrix factorization modules (features)
        """
        if self.masked:
            self._modules.setflags(write=True)
            self._modules.fill(np.nan)
            self._modules[self.mask] = self.w.T
            self._modules.setflags(write=False)
        return self._modules

    @modules.setter
    def modules(self, value):
        if self.masked:
            self._modules.setflags(write=True)
            self._modules[self.mask] = value[self.mask]
            self._modules.setflags(write=False)
            np.copyto(self.w, self._modules[self.mask].T)
        else:
            np.copyto(self._modules, value)
        del self.res

    @property
    def weights(self):
        """
        Recovered weights for all modules and all spikes

        :type: :class:`numpy.ndarray`

        Notes
        -----
        This array is a view of the matrix factorization encodings
        `stnmf.mf.MF.h`.

        See Also
        --------
        ratios : Scalar weight per subunit
        stnmf.mf.MF.h : Matrix factorization weights (encodings)
        """
        return self._weights

    @weights.setter
    def weights(self, value):
        np.copyto(self._weights, value)
        del self.res

    @property
    def subunits(self):
        """
        Localized subunits, i.e. subset of STNMF modules (normalized)

        :type: :class:`numpy.ndarray`
        """
        su = self.modules[self.localized]
        su /= np.linalg.norm(su, axis=(1, 2), keepdims=True)
        return su

    @property
    def ratios(self):
        """
        Scalar weights/contribution ratios of each localized subunit,
        i.e. subset of STNMF weights (averaged and normalized)

        :type: :class:`numpy.ndarray`
        """
        wgt = self.weights[self.localized].mean(axis=1)
        wgt /= np.linalg.norm(wgt)
        return wgt

    @property
    def num_subunits(self):
        """
        Number of localized subunits as determined by their Moran's I
        autocorrelation

        :type: :class:`int`
        """
        return self.localized.size

    """
    Subunit attributes
    ------------------
    """

    @property
    def outlines(self):
        """
        Contour outlines of the localized subunits at contour level
        equivalent to 1.22 standard deviations

        :type: :class:`numpy.ndarray`

        See Also
        --------
        stnmf.spatial.contour : Fit a contour to an image array
        """
        if self._outlines is None:
            self._outlines = spatial.contour(self.subunits, 1.22, smooth=0.5)
            for o in self._outlines:
                o.setflags(write=False)
            self._outlines.setflags(write=False)
        return self._outlines

    @property
    def diameters(self):
        """
        Diameters of the localized subunits as defined by the effective
        diameter of a two-sigma elliptical Gaussian fit in micrometers
        (μm). Only if `pixel_size` is set

        :type: :class:`numpy.ndarray`

        Warns
        -----
        PixelSizeWarning
            If `pixel_size` is not set

        See Also
        --------
        stnmf.spatial.ellipse : Fit a Gaussian ellipse
        pixel_size : Conversion factor to micrometers (μm)
        """
        self._pixel_size_warning()
        if self._diameters is None:
            el = spatial.ellipse(self.subunits)
            a, b = el.T[2:4]
            self._diameters = 2 * np.sqrt(a*b) * self.pixel_size
            self._diameters.setflags(write=False)
        return self._diameters

    @property
    def distances(self):
        """
        Euclidean distances of localized subunits from STA center in
        micrometers (μm). Only if `pixel_size` is set

        Distances are measured from the outline centroids

        :type: :class:`numpy.ndarray`

        Warns
        -----
        PixelSizeWarning
            If `pixel_size` is not set

        See Also
        --------
        outlines : Contour outlines of the localized subunits
        pixel_size : Conversion factor to micrometers (μm)
        """
        self._pixel_size_warning()
        if self._distances is None:
            c_sub = [shapely.LinearRing(o).centroid for o in self.outlines]
            c_sta = shapely.LinearRing(self.sta_outline).centroid
            self._distances = shapely.distance(c_sub, c_sta) * self.pixel_size
            self._distances.setflags(write=False)
        return self._distances

    @property
    def polarities(self):
        """
        Signs of the localized subunits at maximum intensity (positive
        +1 or negative -1)

        :type: :class:`numpy.ndarray`

        See Also
        --------
        flipsign : Flip sign of modules to make weights positive
        """
        if self._polarities is None:
            filters = self.subunits.reshape(self.num_subunits, self.x * self.y)

            # Find maximum intensity pixel
            peak_idx = np.nanargmax(np.abs(filters), axis=1)
            peak = filters[range(self.num_subunits), peak_idx]
            self._polarities = np.copysign(1, peak, dtype=int)

            # Consider non-flipped subunits
            self._polarities *= np.copysign(1, self.ratios, dtype=int)

            self._polarities.setflags(write=False)
        return self._polarities

    @property
    def moransi(self):
        """
        Moran's I autocorrelation of all modules. To obtain the Moran's
        I of the localized subunits only, index the attribute with
        `localized`

        :type: :class:`numpy.ndarray`
        """
        if self._moransi is None:
            self._moransi = spatial.moransi(self.modules)
            self._moransi.setflags(write=False)
        return self._moransi

    @moransi.deleter
    def moransi(self):
        # This attribute is read-only. Delete to force recalculation
        self._reset_localized()

    @property
    def localized(self):
        """
        Indices of localized subunits among all modules as determined by
        their Moran's I autocorrelation and `moransi_threshold`

        :type: :class:`numpy.ndarray`
        """
        return np.nonzero(self.moransi > self.moransi_threshold)[0]

    """
    General parameters
    ------------------
    """

    @property
    def pixel_size(self):
        """
        Pixel size in micrometers (μm) for calculation of spatial
        diameter and distance

        :type: :class:`float`
        """
        return self._pixel_size

    @pixel_size.setter
    def pixel_size(self, um):
        if um < 0:
            warnings.warn('Pixel size cannot be negative. Set to 0 instead.')
            um = 0
        # Reset any attribute based on pixel size
        if getattr(self, '_pixel_size', -1) != um:
            self._reset_sized()
        self._pixel_size = um

    def _pixel_size_warning(self):
        if not self.pixel_size:
            warnings.warn(
                'No `pixel_size` set for micrometer conversion!',
                PixelSizeWarning
            )

    @property
    def moransi_threshold(self):
        """
        Threshold of Morans'I to distinguish localized subunits

        :type: :class:`float`

        See Also
        --------
        stnmf.spatial.moransi : Calculate the autocorrelation
        """
        return self._moransi_threshold

    @moransi_threshold.setter
    def moransi_threshold(self, value):
        if not -1 < value < 1:
            warnings.warn('Threshold must be in interval [-1, 1]. Ignored.')
            return
        # Reset any attribute based on localized subunits
        if getattr(self, '_moransi_threshold', -1) != value:
            self._reset_localized()
        self._moransi_threshold = value

    """
    Further attributes
    ------------------
    """

    @property
    def sta(self):
        """
        Spatial profile of the spike-triggered average (STA)

        :type: :class:`numpy.ndarray`
        """
        return self._sta

    @property
    def sta_outline(self):
        """
        Contour outline of the STA at contour level equivalent to 1.22
        standard deviations.

        :type: :class:`numpy.ndarray`

        See Also
        --------
        stnmf.spatial.contour : Fit a contour to an image array
        """
        if self._sta_outline is None:
            self._sta_outline = spatial.contour(self.sta, 1.22, smooth=1.2)
            self._sta_outline.setflags(write=False)
        return self._sta_outline

    @property
    def sta_diameter(self):
        """
        Diameter of the STA as defined by the effective diameter of a
        two-sigma elliptical Gaussian fit in micrometers (μm). Only if
        `pixel_size` is set

        :type: :class:`float`

        Warns
        -----
        PixelSizeWarning
            If `pixel_size` is not set

        See Also
        --------
        stnmf.spatial.ellipse : Fit a Gaussian ellipse
        pixel_size : Conversion factor to micrometers (μm)
        """
        self._pixel_size_warning()
        if self._sta_diameter is None:
            el = spatial.ellipse(self.sta)
            a, b = el[2:4]
            self._sta_diameter = 2 * np.sqrt(a*b) * self.pixel_size
        return self._sta_diameter

    @property
    def coverage(self):
        """
        Fraction of spatial coverage of localized subunits on the STA

        Coverage is determined from the subunit and STA outlines

        To avoid effects of poorly reconstructed subunits that each
        cover the entire area of the STA, only the non-overlapping
        portions across the subunits are considered. The fraction of
        this non-overlapping subunit area covering the STA area is the
        coverage factor. This measures the intersection while
        encouraging non-overlapping tiling of the subunits.

        :type: :class:`float`
        """
        sta = shapely.Polygon(self.sta_outline)
        sub_all = [shapely.Polygon(o) for o in self.outlines]

        # Only consider the area where subunits do not overlap
        sub_diff = shapely.Polygon(None)
        for i in range(self.num_subunits):
            for j in range(i+1, self.num_subunits):
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore', RuntimeWarning)
                    intersect = sub_all[i].intersection(sub_all[j])
                sub_diff = sub_diff.union(intersect)
        sub_mask = shapely.union_all(sub_all).difference(sub_diff)

        # Fraction of the STA covered by non-overlapping subunit space
        covered = sta.intersection(sub_mask)
        coverage_factor = covered.area / sta.area
        return coverage_factor

    @property
    def modules_info(self):
        """
        Collection of all module properties for all modules for easy
        dot- and multi-indexing

        The attribute is a `numpy.recarray` of size `r` and of the
        following fields. It collects all properties in one combined
        array. This allows to multi-index properties of all or several
        modules at a time, structured in a numpy array.

        :type: :class:`numpy.recarray`

        ================= ==============================================
        :attr:`index`     Module index
        :attr:`module`    Spatial module (not normalized)
        :attr:`weight`    Average contribution weight (not normalized)
        :attr:`moransi`   Moran's I autocorrelation
        :attr:`localized` Whether the modules is considered a subunit
        :attr:`polarity`  Sign of the modules
        :attr:`contour`   Countour outline (for localized only)
        :attr:`diameter`  Diameter (for localized only)
        :attr:`distance`  Distance from STA center (for localized only)
        ================= ==============================================
        """
        dtp = [
            ('index', 'uint16'),
            ('module', self.dtype, (self.x, self.y)),
            ('weight', self.dtype),
            ('moransi', self.dtype),
            ('localized', 'bool'),
            ('polarity', 'int8'),
            ('contour', 'object'),
            ('diameter', 'float32'),
            ('distance', 'float32'),
        ]
        m = np.empty(self.r, dtype=dtp).view(np.recarray)

        loc = self.localized
        m.index = range(self.r)
        m.module = self.modules  # Non-normalized
        m.weight = self.weights.mean(axis=1)  # Non-normalized
        m.moransi = self.moransi
        m.localized[loc] = True
        m.polarity[loc] = self.polarities
        m.contour[loc] = self.outlines
        m.diameter[loc] = self.diameters
        m.distance[loc] = self.distances
        return m


class PixelSizeWarning(UserWarning):
    pass



"""
Visualization (:mod:`stnmf.plot`)
=================================

Collection of functions for visualization of STNMF results. The function
`all` combines all plotting sub-routines for obtaining a figure with all
relevant information at one glance.

.. autosummary::
    :toctree: generated/

    all
    modules
    outlines
    sta
    weights
    autocorrelation
    scalebar
    getcolors

"""
from cycler import Cycler
import numpy as np
from matplotlib import colors as mcolors
from matplotlib.colors import Colormap, LinearSegmentedColormap
from matplotlib.cm import ScalarMappable
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import Divider as mpl_divider
from mpl_toolkits.axes_grid1 import Size as mpl_size
from mpl_toolkits.axes_grid1.anchored_artists import AnchoredSizeBar
import warnings

from .spatial import moransi

__all__ = [
    'all',
    'modules',
    'outlines',
    'sta',
]

# Default colors
default_color = '#1b9e77'
_cmap1 = LinearSegmentedColormap.from_list('rwb', ['red', 'white', 'black'])
_cmap1.set_bad(color='lightseagreen', alpha=0.25)
_cmap2 = plt.get_cmap('binary_r')
_cmap2.name = 'filter'  # Accessing private attribute?
_cmap2.set_bad(color='magenta', alpha=0.25)

# Internal constants
_FIGURE_WIDTH = 6.25  # Fits well on A4 and US letter paper


class ColormapWarning(UserWarning):
    """Turn the breaking error into a friendly warning and continue"""
    pass


try:
    plt.colormaps.register(cmap=_cmap1)
except ValueError as e:
    warnings.warn(str(e), ColormapWarning)
try:
    plt.colormaps.register(cmap=_cmap2, name='filter')
except ValueError as e:
    warnings.warn(str(e), ColormapWarning)


def modules(m, colors=default_color, axs=None, **kwargs):
    """
    Plot the modules in a grid figure

    Parameters
    ----------
    m : (r,x,y) array_like
        Spatial modules, where `r` is the number of modules and `x` and
        `y` are the spatial dimensions

    colors : color, iterable, matplotlib.colors.Colormap, optional
        Single color or color sequence for axis frame to differentiate
        localized modules. If single color, the same color is used on
        the plots of all localized subunits. If iterable or colormap,
        colors are iterated over modules. If None, then no coloring is
        applied. Default is '#1b9e77'

    axs : array_like of matplotlib.axes.Axes, optional
        Existing figure axes to draw the modules into. If None, a new
        figure is created. Default is None

    Keyword arguments
    -----------------
    cmap : str or matplotlib.colors.Colormap, optional
        Color map to use. Recommended are 'rwb' for convenient view
        (red-white-black with masked values lightseagreen) and 'filter'
        for conventional spatial contrast filter colors
        (black-gray-white with masked values magenta). Default is 'rwb'

    localized : (l,) array_like, optional
        Indices of localized modules, where `l` is the number of
        localized modules. Not required, but if already available, may
        speed up figure creation. If None, localized modules are
        determined based on the Moran's I threshold in
        `moransi_threshold`

    moransi_threshold : float, optional
        Moran's I threshold to differentiate localized modules. Default
        is 0.25

    Raises
    ------
    ValueError
        If `m` has incorrect shape.

    IndexError
        If number of `axs` and length `m` do not match.
    """
    m = np.asarray(m)
    if m.ndim != 3:
        raise ValueError('m is expected to be three dimensional')
    r, x, y = m.shape

    if axs is None:
        # Estimate a good axes layout
        nx = int(np.ceil(np.sqrt(r)))
        ny = int(np.ceil(r / nx))
        fig_ratio = (nx * x) / (ny * y)
        fig_width = _FIGURE_WIDTH
        fig_height = fig_width / fig_ratio
        fig, axs = plt.subplots(ny, nx, figsize=(fig_width, fig_height),
                                sharex=True, sharey=True, squeeze=False)
        fig.subplots_adjust(hspace=0.05, wspace=0.05, left=0.01, right=0.99,
                            bottom=0.01, top=0.99)
        # Remove excess axes
        axs = axs.reshape(-1)
        for i in range(nx*ny - r):
            axs[r + i].remove()
    else:
        axs = np.atleast_1d(axs).reshape(-1)
        if axs.size != r:
            raise IndexError('Number of axes and modules inconsistent')

    cmap = plt.get_cmap(kwargs.get('cmap', 'rwb') or 'rwb')
    im_prop = dict(origin='lower', interpolation='nearest', cmap=cmap)
    plt.setp(axs, xticks=[], yticks=[], aspect='equal')

    # Flip sign of OFF cells for convenient view (black instead of red pixels)
    if cmap.name == 'rwb' and not np.any(m > 0):
        m = -m

    for i, (module, ax) in enumerate(zip(m, axs)):
        vm = np.nanmax(np.abs(module), initial=0) or 1
        ax.imshow(module.T, vmin=-vm, vmax=vm, **im_prop)
        plt.setp(ax.spines.values(), lw=1, color='#404040')

    # Color the localized modules
    if colors is not None:
        localized = kwargs.get('localized', None)
        if localized is None:
            thr = kwargs.get('moransi_threshold', 0.25)
            localized = np.nonzero(moransi(m) > thr)[0]
        colors = getcolors(colors, len(localized))
        for ax, sty in zip(axs[localized], colors()):
            plt.setp(ax.spines.values(), lw=2, **sty)


def outlines(subunits, sta=None, colors=default_color, numbered=True,
             filled=True, ax=None, **kwargs):
    """
    Plot the outlines of the modules

    Parameters
    ----------
    subunits : (r,) array_like
        Outlines of `r` subunits. Each element is an `(2, n)` array_like
        containing `n` coordinate pairs. `n` may be different for each
        outline. Outlines may be arbitrary, i.e. either contours or
        ellipses

    sta : (n,2) array_like, optional
        Outline of spike-triggered average with `n` coordinate pairs.
        Omit to only draw the subunit outlines. Default is None

    colors : color, iterable, matplotlib.colors.Colormap, optional
        Single color or color sequence for the subunit outlines. If
        single color, the same color is used for all outlines. If
        iterable or colormap, colors are iterated over outlines. Default
        is '#1b9e77'

    numbered : bool, optional
        Number the subunit outlines. Numbering is done in the order of
        the subunits provided. Default is True

    filled : bool, opional
        Color fill the subunits outlines. Default is True

    ax : matplotlib.axes.Axes, optional
        Existing figure axis to draw the outlines into. If None, a new
        figure is created. Default is None

    Notes
    -----
    There is no error handling in favor of performance
    """
    if ax is None:
        fig_width = fig_height = 2
        fig, ax = plt.subplots(figsize=(fig_width, fig_height))
        fig.subplots_adjust(left=0.01, right=0.99, bottom=0.01, top=0.99)
        ax.set(facecolor='none', frame_on=False)

    ax.set(xticks=[], yticks=[], aspect='equal')
    plt.setp(ax.spines.values(), lw=1)
    fprop = dict(ha='center', va='center', clip_on=False, size=12, zorder=4)
    lw = 2

    # Subunit outlines
    colors = getcolors(colors, len(subunits))
    fcolors = getcolors(colors, len(subunits), key='facecolor')
    multi_colored = len(colors) > 1
    sty = dict(clip_on=False)
    sty2 = dict(solid_joinstyle='round', solid_capstyle='round', clip_on=False)
    for i, (outline, col, fc) in enumerate(zip(subunits, colors(), fcolors())):
        outline_t = np.transpose(outline)
        if filled and multi_colored:
            # All outlines different color: Colored outlines
            ax.plot(*outline_t, zorder=3, lw=lw, **col, **sty2)
            ax.fill(*outline_t, zorder=1, ec='none', alpha=0.2, **fc, **sty)
        elif filled:
            # All outlines one color: Black outline, more opaque fill
            ax.plot(*outline_t, zorder=3, lw=lw, color='black', **sty2)
            ax.fill(*outline_t, zorder=1, ec='none', alpha=0.3, **fc, **sty)
        else:
            ax.plot(*outline_t, zorder=3, lw=lw, **col, **sty2)
        if numbered:
            txy = np.mean(outline, 0)
            col = mcolors.to_rgb(col['color'])
            col = mcolors.rgb_to_hsv(col)
            col[2] = min(0.8, col[2])  # Darken color for readability
            col = mcolors.hsv_to_rgb(col)
            ax.text(*txy, f'{i+1:d}', **fprop, color=col)

    # STA outline
    if sta is not None:
        ax.plot(*np.transpose(sta), zorder=2, lw=lw, color='#4c4c4c',
                clip_on=False)


def sta(s, ax=None, **kwargs):
    """
    Plot the spike-triggered average (STA)

    Parameters
    ----------
    s : (x,y) array_like
        Spatial profile of the spike-triggered average

    ax : matplotlib.axes.Axes, optional
        Existing figure axes to draw the STA into. If None, a new figure
        is created. Default is None

    Keyword arguments
    -----------------
    cmap : str or matplotlib.colors.Colormap, optional
        Color map to use. Recommended are 'filter' for conventional
        spatial contrast filter colors (black-gray-white with masked
        values magenta) and 'rwb' for convenient view (red-white-black
        with masked values lightseagreen). Default is 'filter'

    Raises
    ------
    ValueError
        If `s` has incorrect shape.
    """
    s = np.asarray(s)
    if s.ndim != 2:
        raise ValueError('s is expected to be two dimensional')
    x, y = s.shape

    if ax is None:
        fig_ratio = x / y
        fig_width = 2
        fig_height = fig_width / fig_ratio
        fig, ax = plt.subplots(figsize=(fig_width, fig_height))
        fig.subplots_adjust(left=0.01, right=0.99, bottom=0.01, top=0.99)
        ax.set_facecolor('none')

    cmap = plt.get_cmap(kwargs.get('cmap', 'filter') or 'filter')
    im_prop = dict(origin='lower', interpolation='nearest', cmap=cmap)
    ax.set(xticks=[], yticks=[], aspect='equal')
    plt.setp(ax.spines.values(), lw=1, color='#404040')

    vm = np.nanmax(np.abs(s), initial=0) or 1
    ax.imshow(s.T, vmin=-vm, vmax=vm, **im_prop)


def weights(w, m, colors=default_color, ax=None, yticks=True, **kwargs):
    """
    Plot the average weights of the modules

    Parameters
    ----------
    w : (r,) or (r,sp) array_like
        Weight or weights of each module. Either already averaged `(r,)`
        or all weights `(r, sp)`, where `r` is the number of modules and
        `sp` is the number of spikes

    m : (l,) or (r,x,y) array_like
        Indices of localized modules or the spatial modules themselves,
        where `l` is the number of localized modules or `r` is the
        number of all modules and `x` and `y` are the spatial
        dimensions. If modules, localized modules are determined based
        on the Moran's I threshold in `moransi_threshold`

    colors : color, iterable, matplotlib.colors.Colormap, optional
        Single color or color sequence for the data points to
        differentiate localized modules. If single color, the same color
        is used on the plots of all localized subunits.  If iterable or
        colormap, colors are iterated over modules. Default is '#1b9e77'

    ax : matplotlib.axes.Axes, optional
        Existing figure axes to draw the weights into. If None, a new
        figure is created. Default is None

    yticks : bool, optional
        Display weights on y-axis. Default True

    Keyword arguments
    -----------------
    moransi_threshold : float, optional
        Moran's I threshold to differentiate localized modules. Default
        is 0.25

    Raises
    ------
    ValueError
        If `m` or `w` has incorrect shape.
    """
    w = np.squeeze(w)
    if w.ndim == 2:
        w = w.mean(axis=-1)
    if w.ndim != 1:
        raise ValueError('w is expected to be one or two dimensional')
    r = w.size

    ndim = np.ndim(m)
    if ndim == 1:
        thr = np.isin(range(r), m)
    elif ndim == 3:
        thr = moransi(m) > kwargs.get('moransi_threshold', 0.25)
        m = np.nonzero(thr)[0]
    else:
        raise ValueError('m is expected to be one or three dimensional')
    colors = getcolors(colors, len(m), key='mfc')

    if ax is None:
        fig_ratio = 4
        fig_width = 3
        fig_height = fig_width / fig_ratio
        fig, ax = plt.subplots(figsize=(fig_width, fig_height))
        fig.subplots_adjust(left=0.01, right=0.99, bottom=0.01, top=0.99)
        ax.set_facecolor('none')

    ax.set_frame_on(False)
    mprop = dict(mew=1, ms=5, clip_on=False)
    ind = np.arange(r, dtype=int)
    ax.plot(ind[~thr], w[~thr], 'ko', mfc='#a0a0a0', **mprop)
    for x, y, sty in zip(ind[thr], w[thr], colors()):
        ax.plot(x, y, 'ko', **mprop, **sty)

    ax.set_xticks([])
    if not yticks:
        ax.set_yticks([0])
        ax.tick_params(axis='y', pad=-5)
    ax.tick_params(length=0)
    plt.setp(ax.get_yticklabels(), 'fontsize', 8, 'family',
             plt.rcParams['font.family'])
    ax.grid(axis='y', color='lightgray')
    ax.get_ygridlines()[0].set_xdata((0.05, 1))
    margin = np.ptp(w) * 0.1
    ax.set_ylim(min(w.min(), 0) - margin, max(w.max(), 0) + margin)


def autocorrelation(m, colors=default_color, ax=None, **kwargs):
    """
    Plot the spatial autocorrelation of the modules

    Parameters
    ----------
    m : (r,) or (r,x,y) array_like
        Moran's I of the spatial modules, or the modules themselves,
        where `r` is the number of modules `x` and `y` are the spatial
        dimensions. If modules, the autocorrelation is calculated using
        Moran's I

    colors : color, iterable, matplotlib.colors.Colormap, optional
        Single color or color sequence for the data points to
        differentiate localized modules. If single color, the same color
        is used on the plots of all localized subunits. If iterable or
        colormap, colors are iterated over modules. Default is '#1b9e77'

    ax : matplotlib.axes.Axes, optional
        Existing figure axes to draw the autocorrelation into. If None,
        a new figure is created. Default is None

    Keyword arguments
    -----------------
    moransi_threshold : float, optional
        Moran's I threshold to differentiate localized modules. Default
        is 0.25
    """
    ndim = np.ndim(m)
    if ndim == 3:
        m = moransi(m)
    elif ndim != 1:
        raise ValueError('m is expected to be one or three dimensional')
    m = np.clip(m, 0, 1)
    r = m.size
    thr = m > kwargs.get('moransi_threshold', 0.25)
    num_loc = np.count_nonzero(thr)
    colors = getcolors(colors, num_loc, key='mfc')

    if ax is None:
        fig_ratio = 4
        fig_width = 3
        fig_height = fig_width / fig_ratio
        fig, ax = plt.subplots(figsize=(fig_width, fig_height))
        fig.subplots_adjust(left=0.01, right=0.99, bottom=0.01, top=0.99)
        ax.set_facecolor('none')

    mprop = dict(mew=1, ms=5, clip_on=False)
    ind = np.arange(r, dtype=int)
    ax.plot(ind[~thr], m[~thr], 'ko', mfc='#a0a0a0', **mprop)
    for x, y, sty in zip(ind[thr], m[thr], colors()):
        ax.plot(x, y, 'ko', **mprop, **sty)

    ax.set(xticks=[], yticks=[0, 1], ylim=[-0.075, 1.05])
    ax.set_yticks([0.5], minor='True')
    ax.tick_params(axis='x', length=0)
    ax.tick_params(axis='y', length=3.5, direction='in', pad=-12, which='both')
    ax.tick_params(axis='y', length=1.75, which='minor')
    plt.setp(ax.spines.values(), 'visible', False)
    plt.setp(ax.spines['left'], 'visible', True, 'bounds', [0, 1])
    plt.setp(ax.get_yticklabels(), 'fontsize', 8, 'family',
             plt.rcParams['font.family'])


def all(s, m, s_outl, m_outl, w=None, autocorr=None,  # noqa: C901
        pixel_size=None, colors='hsv', **kwargs):
    """
    Create figure combining the modules, their outlines, the
    spike-triggered average (STA), and - optionally - the weights and
    autocorrelation values

    Parameters
    ----------
    s : (x,y) array_like
        Spatial spike triggered average (STA), where `x` and `y` are the
        spatial dimensions

    m : (r,x,y) array_like
        Spatial modules, where `r` is the number of modules and `x` and
        `y` are the spatial dimensions

    s_outl : (n,2) array_like
        Outline of STA with `n` coordinate pairs. Default is None

    m_outl : (l,) array_like
        Outlines of the `l` localized subunits. Each element is an
        (2, n) array_like containing `n` coordinate pairs. `n` may be
        different for each outline. Outlines may be arbitrary, i.e.
        either contours or ellipses.

    w : (r,) or (r,sp) array_like, optional
        Weight or weights of each module. Either already averaged `(r,)`
        or all weights `(r, sp)`, where `r` is the number of modules and
        `sp` is the number of spikes. If None, weights and
        autocorrelation plots are omitted (reduced figure). Default is
        None

    autocorr : (r,) array_like, optional
        Moran's I of each spatial modules. Not required, but if already
        available, may speed up figure creation. Default is None

    pixel_size : float, optional
        Size of a stimulus pixel (stixel) in micrometers (μm) for
        displaying the scale bar. If not specified, no scale bar is
        shown. Default is None

    colors : color, iterable, matplotlib.colors.Colormap, optional
        Single color or color sequence to differentiate localized
        modules. If single color, the same color is used on the plots of
        all localized subunits. If iterable or colormap, colors are
        iterated over plots of localized subunits. Default is 'hsv'

    Keyword arguments
    -----------------
    cmap : str or matplotlib.colors.Colormap, optional
        Override colormaps for both STA and modules. By default the
        modules and STA plots have two different colormaps. If
        specified, the colormap is used for all subplots. Recommended
        are 'rwb' for convenient view (red-white-black with masked
        values lightseagreen) and 'filter' for conventional spatial
        contrast filter colors (black-gray-white with masked values
        magenta). If None, modules are visualized with 'rwb' and STA
        with 'filter'. Default is None

    moransi_threshold : float, optional
        Moran's I threshold to differentiate localized modules. Default
        is 0.25

    Returns
    -------
    fig : matplotlib.figure.Figure
        Created figure

    Raises
    ------
    ValueError
        If `s`, `m`, `outl`, or `w` have incorrect shape.

    IndexError
        If the spatial dimensions of `s` and `m` do not match.
    """
    s = np.asarray(s)
    if s.ndim != 2:
        raise ValueError('s is expected to be two dimensional')

    m = np.asarray(m)
    if m.ndim != 3:
        raise ValueError('m is expected to be three dimensional')
    r, x, y = m.shape

    if (x, y) != s.shape:
        raise IndexError('Spatial dimensions of STA and modules do not match')

    if w is not None:
        w = np.asarray(w)
        if w.ndim == 2:
            w = w.mean(axis=1)
        elif w.ndim != 1:
            raise ValueError('w is expected to be one or two dimensional')
        if w.size != r:
            raise IndexError('Number of modules and weights do not match')
        full = True
    else:
        full = False

    if autocorr is not None:
        autocorr = np.asarray(autocorr)
        if autocorr.ndim != 1:
            raise ValueError('autocorr is expected to be one dimensional')
        if autocorr.size != r:
            raise IndexError('autocorr has wrong length')
    else:
        autocorr = moransi(m)

    # Get localized modules and corresponding colors
    thr = kwargs.setdefault('moransi_threshold', 0.25)
    kwargs['localized'] = loc = np.nonzero(autocorr > thr)[0]
    kwargs['colors'] = getcolors(colors, len(loc))
    cmap = kwargs.get('cmap', None)

    # Create figure layout with exact and identical axes padding/gaps
    fig_width = _FIGURE_WIDTH
    pad_edge_inch = 0.05
    pad_gap_inch = 0.05
    pad_gaph_inch = 0.025
    pad_gap2_inch = 0.1
    pad_gap3_inch = 0.35
    nx = int(np.ceil(np.sqrt(r)))
    ny = int(np.ceil(r / nx))
    px_aspect = y/x

    # Calculate ratio between module and sta plot sizes
    vertl_size = mpl_size.Add(mpl_size.Fixed(pad_gap_inch * max(ny-2, 0)),
                              mpl_size.Scaled(y * ny))

    # Fixed sizes in inches
    sz_edge = mpl_size.Fixed(pad_edge_inch)
    sz_gap = mpl_size.Fixed(pad_gap_inch)
    sz_gaph = mpl_size.Fixed(pad_gaph_inch)
    sz_gap2 = mpl_size.Fixed(pad_gap2_inch)
    sz_gap3 = mpl_size.Fixed(pad_gap3_inch)
    sz_nop = mpl_size.Fixed(0)

    # Subplot sizes
    sz_mod_x = mpl_size.Scaled(x)
    sz_mod_y = mpl_size.Scaled(y)
    sz_sta_y = mpl_size.Fraction(0.5, vertl_size)
    sz_sta_x = mpl_size.Fraction(1/px_aspect, sz_sta_y)
    sz_cbar_y = sz_gap
    sz_cbar_x = mpl_size.Fraction(0.5, sz_sta_x)
    sz_cbars_y_neg = mpl_size.Fixed(-pad_gap_inch*2-pad_gaph_inch)
    sz_sta_y2 = mpl_size.Add(sz_sta_y, sz_cbars_y_neg)
    sz_ext_x = mpl_size.Fixed((fig_width - (pad_edge_inch*2+pad_gap3_inch))/2)
    sz_ext_y = mpl_size.Fraction(0.3, sz_ext_x)

    # Axes segregation
    horzt = [sz_edge] + [sz_mod_x, sz_gap]*nx + [sz_sta_x, sz_edge]
    horzc = [sz_edge] + [sz_mod_x, sz_gap]*nx + [sz_cbar_x, sz_cbar_x, sz_edge]
    horzb = [sz_edge, sz_ext_x, sz_gap3, sz_ext_x, sz_edge]
    vertb = [sz_edge, sz_ext_y, sz_gap2] if full else [sz_edge, sz_nop, sz_nop]
    vertl = vertb + list(np.tile([sz_mod_y, sz_gap], ny)[:-1]) + [sz_edge]
    vertr = vertb + [sz_sta_y2, sz_cbar_y, sz_gaph, sz_cbar_y, sz_gap,
                     sz_sta_y, sz_edge]

    # Create figure and axes
    ax_prop = dict(xticks=[], yticks=[], facecolor='none')
    fig, axs = plt.subplots(r+6 if full else r+4, subplot_kw=ax_prop)

    rect = (0, 0, 1, 1)
    dividerl = mpl_divider(fig, rect, horzt, vertl, aspect=True)
    dividerr = mpl_divider(fig, rect, horzt, vertr, aspect=True)
    dividerc = mpl_divider(fig, rect, horzc, vertr, aspect=True)
    dividerb = mpl_divider(fig, rect, horzb, vertr, aspect=False)

    # Adjust figure size
    sw = np.sum(dividerl.get_horizontal_sizes(fig.canvas), axis=0)
    sh = np.sum(dividerl.get_vertical_sizes(fig.canvas), axis=0)
    w_r = (fig_width - sw[1]) / sw[0]  # Ratio of scaled and fixed sizes
    fig_h = w_r * sh[0] + sh[1]
    fig.set_size_inches(fig_width, fig_h)

    # Arrange subplots
    for i in range(r):
        py, px = divmod(i, nx)
        py = ny-1 - py
        axs[i].set_axes_locator(dividerl.new_locator(px*2+1, 2+py*2+1))
    last_col = nx*2 + 1
    c_end = last_col+2 if pixel_size is None else None
    axs[r+0].set_axes_locator(dividerr.new_locator(last_col, 8))
    axs[r+1].set_axes_locator(dividerr.new_locator(last_col, 3))
    axs[r+2].set_axes_locator(dividerc.new_locator(last_col, 6, c_end))
    axs[r+3].set_axes_locator(dividerc.new_locator(last_col, 4, c_end))
    if full:
        axs[r+4].set_axes_locator(dividerb.new_locator(1, 1))
        axs[r+5].set_axes_locator(dividerb.new_locator(3, 1))

    with plt.rc_context({'font.family': 'Helvetica'}):  # Adjust font

        # Plot subunits
        modules(m, axs=axs[:r], **kwargs)

        # Plot ST
        sta(s, ax=axs[r], **kwargs)

        # Plot outlines
        ax = axs[r+1]
        outlines(m_outl, s_outl, ax=ax, **kwargs)
        ax.set_frame_on(False)
        # ax.set(xlim=[-0.5, x-0.5])  # Scale it to the STA
        ax.margins(y=0.15)  # Auto-scale with some margins

        # Add scale bar
        if pixel_size is not None:
            bar_lengths = np.array([25] + list(range(50, 450, 50)) + [np.inf])
            reasonable_length = x*pixel_size // 2.5
            idx = np.searchsorted(bar_lengths, reasonable_length, side='left')
            scalebar(axs[r], pixel_size, extent=bar_lengths[max(idx-1, 0)],
                     loc='upper right', bbox_to_anchor=[1, 0], label_top=False,
                     pad=pad_gap_inch * 7.2, borderpad=0, sep=2)

        # Add color bar(s)
        cmap_prop = dict(orientation='horizontal', ticks=[])
        mapl = ScalarMappable(norm=None, cmap=cmap or 'filter')
        fig.colorbar(mapl, cax=axs[r+2], **cmap_prop)
        if cmap is not None:
            fig.delaxes(axs[r+3])
        else:
            mapl = ScalarMappable(norm=None, cmap='rwb')
            fig.colorbar(mapl, cax=axs[r+3], **cmap_prop)

        # Extended figure
        if full:
            # Plot weights
            ax = axs[r+4]
            weights(w, loc, ax=ax, yticks=False, **kwargs)
            ax.text(0.99, 1, 'Average weight', ha='right', va='top',
                    fontsize=10, transform=ax.transAxes)

            # Plot autocorrelations
            ax = axs[r+5]
            autocorrelation(autocorr, ax=ax, **kwargs)
            ax.text(0.99, 1, 'Autocorrelation', ha='right', va='top',
                    fontsize=10, transform=ax.transAxes)

    return fig


def scalebar(ax, factor, extent=100, unit='μm', **kwargs):
    """
    Wrapper function for adding a scale bar to an axes

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        Figure axes

    factor : float
        Scaling factor from data coordinates to units

    extent : float, optional
        Length of the scale bar in units. Default is 100

    unit : str, optional
        Unit to be displayed. Default is 'μm'

    Keyword arguments
    -----------------
    kwargs
        Additional arguments to be passed to
        `mpl_toolkits.axes_grid1.anchored_artists.AnchoredSizeBar`
    """
    if factor == 0:
        return
    size = extent / factor
    text = f'{extent:1.4g} {unit}'

    dd = dict(color='black', label_top=True, sep=1.5, pad=0.15, frameon=False,
              size_vertical=0.1, loc='lower left', bbox_transform=ax.transAxes)
    dd.update(kwargs)

    asb = AnchoredSizeBar(ax.transData, size, text, **dd)
    asb.set(clip_on=False)

    # Change cap style for accurate depiction
    rect = asb.size_bar.get_children()[0]
    rect.set(capstyle='butt', joinstyle='miter')
    ax.add_artist(asb)
    return asb


def getcolors(colors, num, key='color'):
    """
    Turn an arbitrary description of colors into a cycler

    Parameters
    ----------
    colors : color, str, iterable, or matplotlib.colors.Colormap
        Single color or iterable (list, tuple, array_like) of colors in
        any representation, or color map. If str, registered colormaps
        will be searched in matplotlib, otherwise it is assumed a named
        color or if 'auto' the default color cycle is used

    num : int
        Number of colors requested. Colors will repeat after that number

    key : str
        Specify the key of the prop cycle. Default is 'color'

    Returns
    -------
    cycler : cycler.Cycler
        Color cycler with the key word specified with `key`
    """
    num = max(1, num)
    if isinstance(colors, Cycler):
        colors = plt.cycler(colors)
        colors.change_key(colors.keys.pop(), key)
        return colors
    if isinstance(colors, str):
        if colors == 'auto':
            colors = plt.cycler(plt.rcParams['axes.prop_cycle'])
            colors.change_key('color', key)
            return colors
        elif colors in plt.colormaps():
            colors = plt.get_cmap(colors)
    if isinstance(colors, Colormap):
        colors = colors(np.linspace(0, 1, min(colors.N+1, num+1))[:-1])
    else:
        colors = mcolors.to_rgba_array(colors)[:num]
    return plt.cycler(key, colors)



"""
Spike-triggered non-negative matrix factorization
=================================================

A fast and versatile implementation of spike-triggered non-negative
matrix factorization (STNMF) based on accelerated fast hierarchical
alternating least squares (AF-HALS) algorithms.

This Python package allows fast inference of receptive-field subunits
from the spiking responses of retinal ganglion cells including methods
of hyperparameter tuning.

Described in the paper:

    |citation-apa|

..
    See documentation for citation:
        https://stnmf.readthedocs.io
    If you use this software, please cite it as indicated.

The package provides code to recover spatial subunits from a
spike-triggered stimulus ensemble using matrix factorization to solve

    :math:`\\mathbf{V} \\approx \\mathbf{W} \\mathbf{H} \\,,`

with a non-negativity constraint on :math:`\\mathbf{W}`. Here
:math:`\\mathbf{V}` is the effective spike-triggered stimulus ensemble
(STE), an `(n, m)` :term:`array_like` consisting of `n` pixels and `m`
spikes and :math:`\\mathbf{W}` is an `(n, r)` :term:`array_like`
containing the recovered spatial modules with the corresponding weights
:math:`\\mathbf{H}` as `(r, m)` :term:`array_like`.
"""
__version__ = '1.1.0'
__all__ = [
    'STNMF',
    'preprocessing',
    'spatial',
    'plot',
    'init',
    'nnls',
    'callbacks',
]

from . import spatial
from . import preprocessing
from . import plot
from . import init
from . import nnls
from . import callbacks
from .core import STNMF



"""
Preprocessing routines (:mod:`stnmf.preprocessing`)
===================================================

Data is brought into appropriate format to be used with and passed to
:class:`stnmf.STNMF`.

.. automodule:: stnmf.preprocessing.spiketriggered
  :no-members:
  :no-inherited-members:
  :no-special-members:

"""
__all__ = [
    'spiketriggered'
]

from . import spiketriggered



"""
Nonnegative least squares solvers (:mod:`stnmf.nnls`)
=====================================================

Functions that solve or approximate the non-negative least squares
problem which sits at the core of non-negative matrix factorization.

.. autosummary::
    :toctree: generated/

    afhals

"""
import numpy as np

__all__ = [
    'afhals',
]


def afhals(vht, hht, w0, rho=0.0, alpha=0.5, eps=0.1, lmbda=0, inplace=False):
    """
    Accelerated fast hierarchical alternating least squares (AF-HALS)
    update

    AF-HALS is based on fast hierarchical alternating least squares
    (Fast HALS) [1]_ combined with accelerated hierarchical alternating
    least squares (A-HALS) [2]_, as inner update for `W` in:

        :math:`\\mathbf{V} \\approx \\mathbf{W}\\mathbf{H}^\\top`

    Parameters
    ----------
    vht : (n,r) array_like
        Product of `v` and `h.T`, where `v` is an `(n, m)` array_like
        and `h` is an `(r, m)` array_like

    hht : (r,r) array_like
        Product of `h` and `h.T`, where `h` is `(r, m)`

    w0 : (m,r) array_like
        Initialized matrix `w`

    rho : float, optional
        Rho for `w` as computed in [2]_. If zero, no inner update loops
        will be done, which reverts back to fast HALS. Default is 0.0

    alpha : float, optional
        This is the scaling parameter for rho as described in [2]_.
        Default is 0.5

    eps : float, optional
        Epsilon is the scaling parameter for the relative error as
        described in [2]_. Default is 0.1

    lmbda : float, optional
        Weight of sparsity constraint. Default is 0

    inplace : bool, optional
        If True, compute `w` in-place by modifying `w0`. This only
        succeeds if `w` is of type `numpy.ndarray`. Default is False

    Returns
    -------
    w : (m,r) numpy.ndarray
        Updated matrix `w`

    Notes
    -----
    Since the update is symmetric, if updating `h` instead, the first
    three parameters change to the following. Additionally, a suitable
    `rho` should be supplied for `h` instead.

    Other Parameters
    ----------------
    vtw : (m,r) array_like
        Product of `v.T` and `w`, where `v` is an `(n, m)` array_like
        and `w` is an `(n, r)` array_like

    wtw : (r,r) array_like
        Product of `w.T` and `w`, where `w` is an `(n, r)` array_like

    ht0 : (m,r) array_like
        Initialized matrix `h.T`, where `h` is an `(r, m)` array_like

    Notes
    -----
    All of the matrix operations are performed in-place to reduce memory
    cost and computational time on repeated allocations. The update of
    `w` is performed with

    .. math::
        \\mathbf{W}_j\\leftarrow\\left [\\mathbf{W}_j+
        (\\mathbf{V}\\mathbf{H}^\\top)_j-\\mathbf{W}
        (\\mathbf{H}\\mathbf{H}^\\top)_j-\\lambda\\right]_+\\,,

    where :math:`\\mathbf{A}_j` is the :math:`j` th column of
    :math:`\\mathbf{A}` and :math:`[\\mathbf{A}]_+` is the element wise
    :math:`\\max(\\mathbf{A}, 0)`.

    The expensive matrix calculations
    :math:`\\mathbf{V}\\mathbf{H}^\\top` and
    :math:`\\mathbf{H}\\mathbf{H}^\\top` are done separately from the
    outer update loop as suggested by [2]_. Additionally,
    :math:`\\mathbf{W}+(\\mathbf{V}\\mathbf{H}^\\top)` is
    calculated only once for each inner update loop.

    References
    ----------
    .. [1] Cichocki, A., & Phan, A. H. (2009). Fast local algorithms
           for large scale nonnegative matrix and tensor
           factorizations. IEICE Transactions on Fundamentals of
           Electronics, Communications and Computer Sciences, E92-A(3),
           708–721. https://doi.org/10.1587/transfun.E92.A.708

    .. [2] Gillis, N., & Glineur, F. (2012). Accelerated Multiplicative
           Updates and Hierarchical ALS Algorithms for Nonnegative
           Matrix Factorization. Neural Computation, 24(4), 1085–1105.
           https://doi.org/10.1162/NECO_a_00256
    """
    vht = np.asarray(vht)
    hht = np.asarray(hht)
    w = np.array(w0, copy=not inplace)
    n, r = w.shape

    max_iter = int(1 + alpha * rho)
    w_prev = w.copy()
    err_0 = None

    # Pre-allocate memory for in-place operations
    mem_n = np.empty((n,), dtype=w.dtype)
    mem_nr = np.empty_like(w)

    for i in range(max_iter):
        # Inner update loop
        np.add(w, vht, out=mem_nr)
        for j in range(r):
            # Opt: np.maximum(w[:, j] + vht[:, j] - w @ hht[:, j] - lmda, 0)
            np.matmul(w, hht[:, j], out=mem_n)
            np.subtract(mem_nr[:, j], mem_n, out=w[:, j])
            if lmbda > 0:
                np.subtract(w[:, j], lmbda, out=w[:, j])
            np.maximum(w[:, j], 0, out=w[:, j])

        # Additional termination criterion by improvement
        np.subtract(w, w_prev, out=mem_nr)
        norm = np.linalg.norm(mem_nr, ord='fro')
        if not i:
            err_0 = eps * norm
        elif norm <= err_0:
            break
        np.copyto(w_prev, w)

    # Ensure nonzero columns
    np.nan_to_num(w, nan=1e-16, copy=False)
    w[:, w.sum(axis=0) == 0] = 1e-16

    return w



"""
Example callback to monitor the reconstruction residual
=======================================================
"""
import numpy as np

__all__ = [
    'callback',
]


def callback(self, i, itor, callback_data, interval=100):
    """
    Callback function to view and record the course of the
    reconstruction residual

    The residual is displayed continuously and is afterwards provided as
    a :mod:`numpy.ndarray` stored in the key 'res' in the dictionary
    `callback_data`.

    Parameters
    ----------
    interval : int, optional
        Compute the residual every `interval` iterations. Default is 100

    Notes
    -----
    This callback function is not specific to :class:`stnmf.STNMF` and
    can be used in conjunction with other :mod:`snmt.mf` classes.

    Examples
    --------
    >>> from stnmf import STNMF
    >>> from stnmf.callbacks import residual
    >>> results = dict()
    >>> stnmf = STNMF(ste, callback=residual, callback_data=results,
    ...               callback_kwargs=dict(interval=100))
    >>> print(results['res'])
    """
    if i == 0:
        # Allocate the array to store the values
        callback_data['res'] = np.zeros(self.num_iter//interval+1, self.dtype)

    if i % interval == 0:
        # Add the current residual
        callback_data['res'][i // interval] = self.res

        # View the residual
        if itor.disable:
            print(f'{i:7}it: {self.res:0.4f}')
        else:
            itor.set_postfix(dict(res=f'{self.res:0.4f}'))
